#!/bin/env python3
"""
The metawards command line program.

Usage:
    To get the help for this program and the list of all of the
    arguments (with defaults) use;

    metawards --help
"""

__all__ = ["get_parallel_scheme", "parse_args", "get_hostfile",
           "get_cores_per_node", "get_threads_per_task",
           "scoop_supervisor", "mpi_supervisor",
           "cli"]


def get_parallel_scheme():
    """This function tries to work out which of the different parallel
       methods we should use to distribute work between multiple processes.

       Detected schemes are "mpi4py", "scoop", and if none of these
       are found, then "multiprocessing"
    """
    import sys

    if "mpi4py" in sys.modules:
        return "mpi4py"
    elif "scoop" in sys.modules:
        return "scoop"
    else:
        return "multiprocessing"


def parse_args():
    """Parse all of the command line arguments"""
    import configargparse
    import sys

    metawards_url = "https://metawards.org"

    configargparse.init_argument_parser(
        name="main",
        description=f"MetaWards epidemic modelling - see "
        f"{metawards_url} for more information.",
        prog="metawards")

    parser = configargparse.get_argument_parser("main")

    parser.add_argument('--version', action="store_true",
                        default=None,
                        help="Print the version information about metawards")

    parser.add_argument('-c', '--config', is_config_file=True,
                        help="Config file that can be used to set some "
                             "or all of these command line options.")

    parser.add_argument('-i', '--input', type=str,
                        help="Input file for the simulation that specifies "
                             "the adjustable parameters to change for each "
                             "run of the model. You must supply some "
                             "input to run a model!")

    parser.add_argument('-l', '--line', type=str, default=None, nargs="*",
                        help="Line number (or line numbers) containing the "
                             "values of adjustable parameters to run for this "
                             "run (or runs) of the model If this isn't "
                             "specified then model runs will be performed "
                             "for adjustable parameters given on all lines "
                             "from the input file. You can specify many "
                             "numbers, and ranges are also accepted, e.g. "
                             "'-l 5 6-10 12,13,14' etc. Note that the line "
                             "numbers are 0-indexed, e.g. the first line of "
                             "the file is line 0. Ranges are inclusive, "
                             "so 3-5 is the same as 3 4 5")

    parser.add_argument("-r", '--repeats', type=int, default=None, nargs="*",
                        help="The number of repeat runs of the model to "
                             "perform for each value of the adjustable "
                             "parameters. By default only a single "
                             "run will be performed for each set of "
                             "adjustable parameters. This complements the "
                             "'repeat' column in the input file (in which "
                             "case the repeats are multipled). Also, "
                             "multiple repeat values can be given, in which "
                             "case each value corresponds to a different "
                             "line in the input file")

    parser.add_argument('-s', '--seed', type=int, default=None,
                        help="Random number seed for this run "
                             "(default is to use a random seed)")

    parser.add_argument('-a', '--additional', type=str, default=None,
                        nargs="*",
                        help="File (or files) containing additional "
                             "seed of outbreak for the model. These are "
                             "used to seed additional infections on "
                             "specific days at different locations "
                             "during a model run")

    parser.add_argument('-o', '--output', type=str, default="output",
                        help="Path to the directory in which to place all "
                             "output files (default 'output'). This "
                             "directory will be subdivided if multiple "
                             "adjustable parameter sets or repeats "
                             "are used.")

    parser.add_argument('-d', '--disease', type=str, default=None,
                        help="Name of the disease to model "
                             "(default is 'ncov')")

    parser.add_argument('-m', '--model', type=str, default=None,
                        help="Name of the input model data set for the "
                             "network (default is '2011Data')")

    parser.add_argument('-D', '--demographics', type=str, default=None,
                        help="Name of the demographics file that provides "
                             "information about how a population is modelled "
                             "as multiple demographics. If this is not "
                             "supplied then the population is modelled "
                             "as a single demographic")

    parser.add_argument('--start-date', type=str, default=None,
                        help="Date of the start of the model outbreak. "
                             "This accepts dates either is iso-format, "
                             "or fuzzy dates such as 'monday', 'tomorrow' "
                             "etc. This is used to work out which days "
                             "are weekends, or to make it easier to specify "
                             "time-based events.")

    parser.add_argument('--start-day', type=int, default=0,
                        help="The start day of the model outbreak. By "
                             "default the model outbreak starts on day "
                             "zero (0), with each step of the model "
                             "representing an additional day. Use this "
                             "to start from a later day, which may be "
                             "useful if you want to more quickly reach "
                             "time based events. Note that the passed "
                             "'--start-date' is always day 0, so day 10 "
                             "has a date which is 10 days after start-date")

    parser.add_argument('-p', '--parameters', type=str, default="march29",
                        help="Name of the input parameter set used to "
                             "control the simulation (default 'march29')")

    parser.add_argument('-R', '--repository', type=str, default=None,
                        help="Path to the MetaWardsData repository. If "
                             "unspecified this defaults to the value "
                             "in the environment variable METAWARDSDATA "
                             "or, if that isn't specified, to "
                             "$HOME/GitHub/MetaWardsData",
                             env_var="METAWARDSDATA")

    parser.add_argument('-P', '--population', type=int, default=57104043,
                        help="Initial population (default 57104043)")

    parser.add_argument('-n', '--nsteps', type=int, default=730,
                        help="Maximum number of steps (days) to run for the "
                             "simulation. Each step represents one day in the "
                             "outbreak (default is to run for a maximum "
                             "of two years - 730 days)")

    parser.add_argument('-u', '--user-variables', type=str, default=None,
                        help="Name of the file containing user-defined "
                             "custom variables. These provide extra "
                             "information that can be read by the "
                             "custom integrators or custom extractors.")

    parser.add_argument('--iterator', type=str, default=None,
                        help=f"Name of the iterator to use to advance the "
                             f"outbreak at each step (day). For a full "
                             f"explanation see the tutorial at "
                             f"{metawards_url}")

    parser.add_argument("--extractor", type=str, default=None,
                        help=f"Name of the extractor to use to extract "
                             f"information during a model run. For a full "
                             f"explanation see the tutorial at "
                             f"{metawards_url}")

    parser.add_argument("--mixer", type=str, default=None,
                        help=f"Name of the mixer to use to mix information "
                             f"from multiple demographics together during "
                             f"a model run. For a full explanation see "
                             f"the tutorial at {metawards_url}")

    parser.add_argument("--mover", type=str, default=None,
                        help=f"Name of the mover to use to move the "
                             f"population between demographics during "
                             f"a model run. For a full explanation see "
                             f"the tutorial at {metawards_url}")

    parser.add_argument('--UV', type=float, default=0.0,
                        help="Value for the UV parameter for the model "
                             "(default is 0.0)")

    parser.add_argument('--theme', type=str, default=None,
                        help=f"The theme to use to color the output. "
                             f"Use '--theme simple' if you prefer a "
                             f"simple and colorless output.")

    parser.add_argument('--no-spinner', action="store_true", default=None,
                        help=f"Disable the spinner that spins when little "
                             f"output is being printed to the screen.")

    parser.add_argument("--debug", action="store_true", default=None,
                        help=f"Enable debugging output. This is useful "
                             f"for MetaWards developers or if you are "
                             f"writing your own iterators, extractors etc.")

    parser.add_argument("--debug-level", type=int, default=None,
                        help="Limit debug output to the specified level.")

    parser.add_argument('--nthreads', type=int, default=None,
                        help="Number of threads over which parallelise an "
                             "individual model run. The total number of "
                             "cores used by metawards will be "
                             "nprocesses x nthreads",
                        env_var="OMP_NUM_THREADS")

    parser.add_argument("--nprocs", type=int, default=None,
                        help="The number of processes over which to "
                             "parallelise the different model runs for "
                             "different adjustable parameter sets and "
                             "repeats. By default this will automatically "
                             "work out the number of processes based on "
                             "the way metawards is launched. Use this "
                             "option if you want to specify the number "
                             "of processes manually.")

    parser.add_argument('--hostfile', type=str, default=None,
                        help="The hostfile containing the names of the "
                             "compute nodes over which to run a parallel "
                             "job. If this is not set, the program will "
                             "attempt to automatically get this information "
                             "from the cluster queueing system. Use this "
                             "if the auto-detection fails")

    parser.add_argument('--cores-per-node', type=int, default=None,
                        help="Set the number of processor cores available "
                             "on each of the compute nodes in the cluster "
                             "that will be used to run the models "
                             "(if a cluster is used). If this is not "
                             "set then the program will attempt to "
                             "get this information from the cluster "
                             "queueing system. Use this option if the "
                             "auto-detection fails.")

    parser.add_argument('--auto-bzip', action="store_true",
                        default=None,
                        help="Automatically bz2 compress "
                             "all output files as they are written.")

    parser.add_argument('--no-auto-bzip', action="store_true",
                        default=None,
                        help="Do not automatically bz2 compress "
                             "all output files as they are written.")

    parser.add_argument('--force-overwrite-output', action="store_true",
                        default=False,
                        help="Whether or not to force overwriting of any "
                             "existing output. Using this option will "
                             "tell metawards that it is safe to delete "
                             "the contents of the output directory "
                             "specified in by '--output' if it already "
                             "exists. Dangerous as this can remove "
                             "existing output files")

    parser.add_argument('--max-nodes', type=int, default=16384,
                        help="Maximum number of nodes that can be read")

    parser.add_argument('--max-links', type=int, default=4194304,
                        help="Maximum number of links that can be read")

    parser.add_argument('--profile', action="store_true",
                        default=None, help="Enable profiling of the code")

    parser.add_argument('--no-profile', action="store_true",
                        default=None, help="Disable profiling of the code")

    parser.add_argument('--mpi', action="store_true", default=None,
                        help="Force use of MPI to parallelise across runs")

    parser.add_argument('--scoop', action="store_true", default=None,
                        help="Force use of scoop to parallelise across runs")

    # this hidden option is used to tell the main process started using
    # mpi that it shouldn't try to become a supervisor
    parser.add_argument('--already-supervised', action="store_true",
                        default=None, help=configargparse.SUPPRESS)

    args = parser.parse_args()

    if args.theme:
        from ..utils._console import Console
        Console.set_theme(args.theme)

    if args.no_spinner:
        from ..utils._console import Console
        Console.set_use_spinner(False)

    if args.debug:
        from ..utils._console import Console
        Console.set_debugging_enabled(args.debug, level=args.debug_level)

    if args.version:
        from metawards import print_version_string
        print_version_string()
        sys.exit(0)

    return (args, parser)


def get_hostfile(args):
    """Attempt to find the name of the hostfile used to specify the hosts
       to use in a cluster
    """
    if args.hostfile:
        return args.hostfile

    import os

    # PBS
    hostfile = os.getenv("PBS_NODEFILE")

    if hostfile:
        return hostfile

    # SLURM
    hostfile = os.getenv("SLURM_HOSTFILE")

    if hostfile:
        return hostfile

    return None


def get_cores_per_node(args):
    """Return the number of cores per node in the cluster"""
    if args.cores_per_node:
        return args.cores_per_node

    import os

    try:
        cores_per_node = int(os.getenv("METAWARDS_CORES_PER_NODE"))

        if cores_per_node > 0:
            return cores_per_node

    except Exception:
        pass

    raise ValueError("You must specify the number of cores per node "
                     "using --cores-per-node or by setting the "
                     "environment variable METAWARDS_CORES_PER_NODE")


def get_threads_per_task(args):
    if args.nthreads:
        return args.nthreads

    import os

    try:
        nthreads = int(os.getenv("METAWARDS_THREADS_PER_TASK"))

        if nthreads > 0:
            return nthreads

        nthreads = int(os.getenv("OMP_NUM_THREADS"))

        if nthreads > 0:
            return nthreads
    except Exception:
        pass

    raise ValueError("You must specify the number of threads per task "
                     "using --nthreads or by setting the "
                     "environment variables METAWARDS_THREADS_PER_TASK "
                     "or OMP_NUM_THREADS")


def scoop_supervisor(hostfile, args):
    """Function used by the scoop supervisor to get the information needed to
       form the scoop call to run a scoop version of the program
    """
    import os
    import sys
    from metawards.utils import Console
    Console.print("RUNNING A SCOOP PROGRAM")

    outdir = args.output

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    cores_per_node = get_cores_per_node(args)

    Console.print(
        f"Will run jobs assuming {cores_per_node} cores per compute node")

    # based on the number of threads requested and the number of cores
    # per node, we can work out the number of scoop processes to start,
    # and can write a hostfile that will create the right layout
    nthreads = get_threads_per_task(args)

    Console.print(f"Will use {nthreads} OpenMP threads per model run...")

    tasks_per_node = int(cores_per_node / nthreads)

    Console.print(f"...meaning that the number of model runs per node will be "
                  f"{tasks_per_node}")

    # Next, read the hostfile to get a unique list of hostnames
    hostnames = {}

    with open(hostfile, "r") as FILE:
        line = FILE.readline()
        while line:
            hostname = line.strip()
            hostnames[hostname] = 1
            line = FILE.readline()

    hostnames = list(hostnames.keys())
    hostnames.sort()

    Console.print(f"Number of compute nodes equals {len(hostnames)}")
    Console.print(", ".join(hostnames))

    # how many tasks can we perform in parallel?
    nprocs = tasks_per_node * len(hostnames)

    if args.nprocs:
        if nprocs != args.nprocs:
            Console.warning(
                f"You are using a not-recommended number of "
                f"processes {args.nprocs} for the cluster {nprocs}.")

        nprocs = args.nprocs

    Console.print(
        f"Total number of parallel processes to run will be {nprocs}")
    Console.print(f"Total number of cores in use will be {nprocs*nthreads}")

    # Now write a new hostfile that round-robins the MPI tasks over
    # the nodes for 'tasks_per_node' runs
    hostfile = f"_metawards_hostfile_{os.getpid()}"
    Console.print(f"Writing hostfile to {hostfile}")

    with open(hostfile, "w") as FILE:
        i = 0
        while i < nprocs:
            for hostname in hostnames:
                FILE.write(hostname + "\n")
                i += 1

                if i == nprocs:
                    break

    # now craft the scoop command that will use this hostfile to
    # run the job - remember to pass the option to stop the main process
    # attempting to become a supervisor itself...

    import subprocess
    import shlex

    pyexe = sys.executable
    script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])

    # also need to tell the main program the number of processes
    # as it can't work it out itself
    cmd = f"{pyexe} -m scoop --hostfile {hostfile} -n {nprocs} " \
          f"{script} --already-supervised {args} --nprocs {nprocs}"

    Console.print("Executing scoop job using")
    Console.command(cmd)

    try:
        args = shlex.split(cmd)
        subprocess.run(args).check_returncode()
    except Exception as e:
        Console.error("ERROR: Something went wrong!")
        Console.error(f"{e.__class__}: {e}")
        sys.exit(-1)

    # clean up the hostfile afterwards... (we leave it if something
    # went wrong as it may help debugging)
    os.unlink(hostfile)

    Console.print("Scoop processes completed successfully")


def mpi_supervisor(hostfile, args):
    """Function used by the MPI supervisor to get the information needed to
       form the mpiexec call to run an MPI version of the program
    """
    import os
    import sys
    from metawards.utils import Console

    Console.print("RUNNING AN MPI PROGRAM")

    outdir = args.output

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    cores_per_node = get_cores_per_node(args)

    Console.print(
        f"Will run jobs assuming {cores_per_node} cores per compute node")

    # based on the number of threads requested and the number of cores
    # per node, we can work out the number of mpi processes to start,
    # and can write a hostfile that will create the right layout
    nthreads = get_threads_per_task(args)

    Console.print(f"Will use {nthreads} OpenMP threads per model run...")

    tasks_per_node = int(cores_per_node / nthreads)

    Console.print(f"...meaning that the number of model runs per node will be "
                  f"{tasks_per_node}")

    # Next, read the hostfile to get a unique list of hostnames
    hostnames = {}

    with open(hostfile, "r") as FILE:
        line = FILE.readline()
        while line:
            hostname = line.strip()
            hostnames[hostname] = 1
            line = FILE.readline()

    hostnames = list(hostnames.keys())
    hostnames.sort()

    Console.print(f"Number of compute nodes equals {len(hostnames)}")
    Console.print(", ".join(hostnames))

    # how many tasks can we perform in parallel?
    nprocs = tasks_per_node * len(hostnames)

    if args.nprocs:
        if nprocs != args.nprocs:
            Console.print(f"WARNING: You are using an unrecommended number of "
                          f"processes {args.nprocs} for the cluster {nprocs}.")

        nprocs = args.nprocs

    Console.print(
        f"Total number of parallel processes to run will be {nprocs}")
    Console.print(f"Total number of cores in use will be {nprocs*nthreads}")

    # Now write a new hostfile that round-robins the MPI tasks over
    # the nodes for 'tasks_per_node' runs
    hostfile = f"_metawards_hostfile_{os.getpid()}"
    Console.print(f"Writing hostfile to {hostfile}")

    with open(hostfile, "w") as FILE:
        i = 0
        while i < nprocs:
            for hostname in hostnames:
                FILE.write(hostname + "\n")
                i += 1

                if i == nprocs:
                    break

    # now craft the mpiexec command that will use this hostfile to
    # run the job - remember to pass the option to stop the main process
    # attempt to become a supervisor itself...
    mpiexec = os.getenv("MPIEXEC")

    if mpiexec is None:
        mpiexec = "mpiexec"

    # check for weird mpiexecs...
    import subprocess
    import shlex

    try:
        args = shlex.split(f"{mpiexec} -v")
        p = subprocess.run(args, stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
        v = p.stdout.decode("utf-8").strip()
        Console.print(f"{mpiexec} -v => {v}")

        if v.find("HPE HMPT") != -1:
            raise ValueError(
                "metawards needs a more modern MPI library than HPE's, "
                "so please compile to another MPI and use that.")
    except Exception as e:
        Console.error(f"[ERROR] {e.__class__} {e}")

    pyexe = sys.executable
    script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])

    cmd = f"{mpiexec} -np {nprocs} -hostfile {hostfile} " \
          f"{pyexe} -m mpi4py {script} --already-supervised {args}"

    Console.print("Executing MPI job using")
    Console.command(cmd)

    try:
        args = shlex.split(cmd)
        subprocess.run(args).check_returncode()
    except Exception as e:
        Console.error("ERROR: Something went wrong!")
        Console.error(f"{e.__class__}: {e}")
        sys.exit(-1)

    # clean up the hostfile afterwards... (we leave it if something
    # went wrong as it may help debugging)
    os.unlink(hostfile)

    Console.print("MPI processes completed successfully")


def cli():
    """Main function for the command line interface. This does one of three
       things:

       1. If this is the main process, then it parses the arguments and
          runs and manages the jobs

       2. If this is a worker process, then it starts up and waits for work

       3. If this is a supervisor process, then it query the job scheduling
          system for information about the compute nodes to use, and will then
          set up and run a manager (main) process that will use those
          nodes to run the jobs
    """
    from metawards.utils import Console

    # get the parallel scheme now before we import any other modules
    # so that it is clear if mpi4py or scoop (or another parallel module)
    # has been imported via the required "-m module" syntax
    parallel_scheme = get_parallel_scheme()

    if parallel_scheme == "mpi4py":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        nprocs = comm.Get_size()
        rank = comm.Get_rank()

        if rank != 0:
            # this is a worker process, so should not do anything
            # more until it is given work in the pool
            Console.print(f"Starting worker process {rank+1} of {nprocs-1}...")
            return
        else:
            Console.print("Starting main process...")

    elif parallel_scheme == "scoop":
        Console.print("STARTING SCOOP PROCESS")

    import sys

    args, parser = parse_args()

    if not args.already_supervised:
        hostfile = get_hostfile(args)
        if hostfile:
            # The user has asked to run a parallel job - this means that this
            # process is the parallel supervisor
            if args.mpi:
                mpi_supervisor(hostfile, args)
                return
            elif args.scoop:
                scoop_supervisor(hostfile, args)
                return

            # neither is preferred - if scoop is installed then use that
            try:
                import scoop        # noqa - disable unused warning
                have_scoop = True
            except Exception:
                have_scoop = False

            if have_scoop:
                scoop_supervisor(hostfile, args)
                return

            # do we have MPI?
            try:
                import mpi4py       # noqa - disable unused warning
                have_mpi4py = True
            except Exception:
                have_mpi4py = False

            if have_mpi4py:
                mpi_supervisor(hostfile, args)
                return

            # we don't have any other option, just keep going and
            # use multiprocessing - in this case we don't need a
            # supervisor and this is the main process

    # This is now the code for the main process

    # WE NEED ONE OF these listed options;
    should_run = False

    for arg in [args.input, args.repeats, args.disease, args.additional,
                args.model, args.iterator, args.extractor,
                args.demographics, args.mixer, args.mover]:
        if arg is not None:
            should_run = True
            break

    if not should_run:
        parser.print_help(sys.stdout)
        sys.exit(0)

    if args.repeats is None:
        args.repeats = [1]

    # import the parameters here to speed up the display of help
    from metawards import Parameters, Network, Population, print_version_string

    # print the version information first, so that there is enough
    # information to enable someone to reproduce this run
    print_version_string()

    Console.rule("Initialise")

    if args.input:
        # get the line numbers of the input file to read
        if args.line is None or len(args.line) == 0:
            linenums = None
            Console.print(f"* Using parameters from all lines of {args.input}",
                          markdown=True)
        else:
            from metawards.utils import string_to_ints
            linenums = string_to_ints(args.line)

            if len(linenums) == 0:
                Console.error(f"You cannot read no lines from {args.input}?")
                sys.exit(-1)
            elif len(linenums) == 1:
                Console.print(f"* Using parameters from line {linenums[0]} of "
                              f"{args.input}", markdown=True)
            else:
                Console.print(f"* Using parameters from lines {linenums} of "
                              f"{args.input}", markdown=True)

        from metawards import VariableSets, VariableSet
        variables = VariableSets.read(filename=args.input,
                                      line_numbers=linenums)
    else:
        from metawards import VariableSets, VariableSet
        # create a VariableSets with one null VariableSet
        variables = VariableSets()
        variables.append(VariableSet())

    nrepeats = args.repeats

    if nrepeats is None or len(nrepeats) < 1:
        nrepeats = [1]

    if len(nrepeats) > 1 and len(variables) != len(nrepeats):
        Console.error(f"The number of repeats {len(nrepeats)} must equal the "
                      f"number of adjustable variable lines {len(variables)}")
        raise ValueError("Disagreement in the number of repeats and "
                         "adjustable variables")

    # ensure that all repeats are >= 0
    nrepeats = [0 if int(x) < 0 else int(x) for x in nrepeats]

    if sum(nrepeats) == 0:
        Console.error(f"The number of the number of repeats is 0. Are you "
                      f"sure that you don't want to run anything?")
        raise ValueError("Cannot run nothing")

    if len(nrepeats) == 1 and nrepeats[0] == 1:
        Console.print("* Performing a single run of each set of parameters",
                      markdown=True)
    elif len(nrepeats) == 1:
        Console.print(
            f"* Performing {nrepeats[0]} runs of each set of parameters",
            markdown=True)
    else:
        Console.print(
            f"* Performing {nrepeats} runs applied to the parameters",
            markdown=True)

    variables = variables.repeat(nrepeats)

    # working out the number of processes and threads...
    from metawards.utils import guess_num_threads_and_procs
    (nthreads, nprocs) = guess_num_threads_and_procs(
        njobs=len(variables),
        nthreads=args.nthreads,
        nprocs=args.nprocs,
        parallel_scheme=parallel_scheme)

    Console.print(
        f"\n* Number of threads to use for each model run is {nthreads}",
        markdown=True)

    if nprocs > 1:
        Console.print(f"* Number of processes used to parallelise model "
                      f"runs is {nprocs}", markdown=True)
        Console.print(
            f"* Parallelisation will be achieved using {parallel_scheme}",
            markdown=True)

    # sort out the random number seed
    seed = args.seed

    if seed is None:
        import random
        seed = random.randint(10000, 99999999)

    if seed == 0:
        # this is a special mode that a developer can use to force
        # all jobs to use the same random number seed (15324) that
        # is used for comparing outputs. This should NEVER be used
        # for production code
        Console.warning("Using special mode to fix all random number"
                        "seeds to 15324. DO NOT USE IN PRODUCTION!!!")
    else:
        Console.print(f"* Using random number seed {seed}", markdown=True)

    # get the starting day and date
    start_day = args.start_day

    if start_day < 0:
        raise ValueError(f"You cannot use a start day {start_day} that is "
                         f"less than zero!")

    start_date = None

    if args.start_date:
        try:
            from dateparser import parse
            start_date = parse(args.start_date).date()
        except Exception:
            pass

        if start_date is None:
            from datetime import date
            try:
                start_date = date.fromisoformat(args.start_date)
            except Exception as e:
                raise ValueError(
                    f"Cannot interpret a valid date from "
                    f"'{args.start_date}'. Error is "
                    f"{e.__class__} {e}")

    if start_date is None:
        from datetime import date
        start_date = date.today()

    Console.print(f"* Day zero is {start_date.strftime('%A %B %d %Y')}",
                  markdown=True)

    if start_day != 0:
        from datetime import timedelta
        start_day_date = start_date + timedelta(days=start_day)
        Console.print(f"Starting on day {start_day}, which is "
                      f"{start_day_date.strftime('%A %B %d %Y')}")
    else:
        start_day_date = start_date

    # now find the MetaWardsData repository as this will be needed
    # for the repeat command line too
    (repository, repository_version) = Parameters.get_repository(
        args.repository)

    Console.print(f"* Using MetaWardsData at {repository}", markdown=True)

    if repository_version["is_dirty"]:
        Console.warning("This repository is dirty, meaning that the data"
                        "has not been committed to git. This may make "
                        "this calculation very difficult to reproduce")

    # now work out the minimum command line needed to repeat this job
    args.seed = seed
    args.nprocs = nprocs
    args.nthreads = nthreads
    args.start_date = start_date.isoformat()
    args.repository = repository

    # also print the source of all inputs
    import configargparse
    Console.rule("Souce of inputs")
    p = configargparse.get_argument_parser("main")
    Console.print(p.format_values())

    # print out the command used to repeat this job
    repeat_cmd = "metawards"

    for key, value in vars(args).items():
        if value is not None:
            k = key.replace("_", "-")

            if isinstance(value, bool):
                if value:
                    repeat_cmd += f" --{k}"
            elif isinstance(value, list):
                repeat_cmd += f" --{k}"
                for val in value:
                    v = str(val)
                    if " " in v:
                        repeat_cmd += f" '{v}''"
                    else:
                        repeat_cmd += f" {v}"
            else:
                v = str(value)
                if " " in v:
                    repeat_cmd += f" --{k} '{v}''"
                else:
                    repeat_cmd += f" --{k} {v}"

    Console.rule("Repeating this run")
    Console.print("To repeat this job use the command;")
    Console.command(repeat_cmd)
    Console.print("Or alternatively use the config.yaml file that will be "
                  "written to the output directory and use the command;")
    Console.command("metawards -c config.yaml")

    Console.rule("Parameters")

    # load all of the parameters
    try:
        params = Parameters.load(parameters=args.parameters)
    except Exception as e:
        Console.warning(
            f"Unable to load parameter files. Make sure that you have "
            f"cloned the MetaWardsData repository and have set the "
            f"environment variable METAWARDSDATA to point to the "
            f"local directory containing the repository, e.g. the "
            f"default is $HOME/GitHub/MetaWardsData")
        raise e

    # should we profile the code? (default no as it prints a lot)
    profiler = None

    if args.no_profile:
        profiler = None
    elif args.profile:
        from metawards.utils import Profiler
        profiler = Profiler()

    # load the disease and starting-point input files
    Console.rule("Disease")
    if args.disease:
        params.set_disease(args.disease)
    else:
        params.set_disease("ncov")

    Console.rule("Model data")
    if args.model:
        params.set_input_files(args.model)
    else:
        params.set_input_files("2011Data")

    # load the user-defined custom parameters
    Console.rule("Custom parameters and seeds")
    if args.user_variables:
        custom = VariableSet.read(args.user_variables)
        Console.print(f"Adjusting variables to {custom}")
        custom.adjust(params)
    else:
        Console.print("Not adjusting any parameters...")

    # read the additional seeds
    if args.additional is None or len(args.additional) == 0:
        Console.print("Not using any additional seeds...")
    else:
        for additional in args.additional:
            Console.print(f"Loading additional seeds from {additional}")
            params.add_seeds(additional)

    # extra parameters that are set
    params.UV = args.UV

    # set these extra parameters to 0
    params.static_play_at_home = 0
    params.play_to_work = 0
    params.work_to_play = 0
    params.daily_imports = 0.0

    # the size of the starting population
    population = Population(initial=args.population,
                            date=start_day_date,
                            day=start_day)

    Console.rule("Building the network")
    network = Network.build(params=params,
                            population=population,
                            max_nodes=args.max_nodes,
                            max_links=args.max_links,
                            profiler=profiler)

    if args.demographics:
        from metawards import Demographics
        Console.rule("Specialising into demographics")
        demographics = Demographics.load(args.demographics)
        Console.print(demographics)

        network = network.specialise(demographics,
                                     profiler=profiler,
                                     nthreads=nthreads)

    Console.rule("Preparing to run")
    from metawards import OutputFiles
    from metawards.utils import run_models

    outdir = args.output

    if outdir is None:
        outdir = "output"

    if args.force_overwrite_output:
        prompt = None
    else:
        from metawards import input
        def prompt(x): return input(x, default="y")

    auto_bzip = True

    if args.auto_bzip:
        auto_bzip = True
    elif args.no_auto_bzip:
        auto_bzip = False

    if args.iterator:
        iterator = args.iterator
    else:
        iterator = None

    if args.extractor:
        extractor = args.extractor
    else:
        extractor = None

    if args.mixer:
        mixer = args.mixer
    else:
        mixer = None

    if args.mover:
        mover = args.mover
    else:
        mover = None

    with OutputFiles(outdir, force_empty=args.force_overwrite_output,
                     auto_bzip=auto_bzip, prompt=prompt) as output_dir:
        # write the config file for this job to output/config.yaml
        Console.rule("Running the model")
        CONSOLE = output_dir.open("console.log")
        Console.save(CONSOLE)

        lines = []
        max_keysize = None

        for key, value in vars(args).items():
            if max_keysize is None:
                max_keysize = len(key)
            elif len(key) > max_keysize:
                max_keysize = len(key)

        for key, value in vars(args).items():
            if value is not None:
                key = key.replace("_", "-")
                spaces = " " * (max_keysize - len(key))

                if isinstance(value, bool):
                    if value:
                        lines.append(f"{key}:{spaces} true")
                    else:
                        lines.append(f"{key}:{spaces} false")
                elif isinstance(value, list):
                    s_value = [str(x) for x in value]
                    lines.append(f"{key}:{spaces} [ {', '.join(s_value)} ]")
                else:
                    lines.append(f"{key}:{spaces} {value}")

        CONFIG = output_dir.open("config.yaml", auto_bzip=False)
        lines.sort(key=str.swapcase)
        CONFIG.write("\n".join(lines))
        CONFIG.write("\n")
        CONFIG.flush()
        CONFIG.close()
        lines = None

        result = run_models(network=network, variables=variables,
                            population=population, nprocs=nprocs,
                            nthreads=nthreads, seed=seed,
                            nsteps=args.nsteps,
                            output_dir=output_dir,
                            iterator=iterator,
                            extractor=extractor,
                            mixer=mixer,
                            mover=mover,
                            profiler=profiler,
                            parallel_scheme=parallel_scheme)

        if result is None or len(result) == 0:
            Console.print("No output - end of run")
            return 0

        Console.rule("End of the run", style="finish")

        Console.save(CONSOLE)

    return 0


if __name__ == "__main__":
    import multiprocessing
    multiprocessing.freeze_support()  # needed to stop fork bombs
    cli()




from dataclasses import dataclass as _dataclass
from typing import List as _List

__all__ = ["Disease"]

_default_folder_name = "diseases"


@_dataclass
class Disease:
    """This class holds the parameters about a single disease

       A disease is characterised as a serious of stages, each
       with their own values of the beta, progress, too_ill_to_move
       and contrib_foi parameters. To load a disease use
       the Disease.load function, e.g.

       Examples
       --------
       >>> disease = Disease.load("ncov")
       >>> print(disease)
       Disease ncov
       repository: https://github.com/metawards/MetaWardsData
       repository_branch: master
       repository_version: 0.2.0
       beta = [0.0, 0.0, 0.95, 0.95, 0.0]
       progress = [1.0, 0.1923, 0.909091, 0.909091, 0.0]
       too_ill_to_move = [0.0, 0.0, 0.0, 0.0, 0.0]
       contrib_foi = [1.0, 1.0, 1.0, 1.0, 0.0]
    """
    #: Beta parameter for each stage of the disease
    beta: _List[float] = None
    #: Progress parameter for each stage of the disease
    progress: _List[float] = None
    #: TooIllToMove parameter for each stage of the disease
    too_ill_to_move: _List[float] = None
    #: Contribution to the Force of Infection (FOI) parameter for each
    #: stage of the disease
    contrib_foi: _List[float] = None
    #: Index of the first symptomatic stage
    start_symptom: int = None

    _name: str = None
    _version: str = None
    _authors: str = None
    _contacts: str = None
    _references: str = None
    _filename: str = None
    _repository: str = None
    _repository_version: str = None
    _repository_branch: str = None

    def __str__(self):
        return f"""
* Disease: {self._name}
* loaded from: {self._filename}
* repository: {self._repository}
* repository_branch: {self._repository_branch}
* repository_version: {self._repository_version}
* beta: {self.beta}
* progress: {self.progress}
* too_ill_to_move: {self.too_ill_to_move}
* contrib_foi: {self.contrib_foi}
* start_symptom: {self.start_symptom}
"""

    def __eq__(self, other):
        return self.beta == other.beta and \
            self.progress == other.progress and \
            self.too_ill_to_move == other.too_ill_to_move and \
            self.contrib_foi == other.contrib_foi and \
            self.start_symptom == other.start_symptom

    def __len__(self):
        if self.beta:
            return len(self.beta)
        else:
            return 0

    def N_INF_CLASSES(self):
        """Return the number of stages of the disease"""
        return len(self.beta)

    def _validate(self):
        """Check that the loaded parameters make sense"""
        try:
            n = len(self.beta)

            assert len(self.progress) == n
            assert len(self.too_ill_to_move) == n
            assert len(self.contrib_foi) == n
        except Exception as e:
            raise AssertionError(f"Data read for disease {self._name} "
                                 f"is corrupted! {e.__class__}: {e}")

        if self.start_symptom is None or self.start_symptom < 0 or \
           self.start_symptom >= n:
            raise AssertionError(f"start_symptom {self.start_symptom} is "
                                 f"invalid for a disease with {n} stages")

        self.start_symptom = int(self.start_symptom)

        from .utils._safe_eval import safe_eval_number

        for i in range(0, n):
            try:
                self.progress[i] = safe_eval_number(self.progress[i])
                self.too_ill_to_move[i] = safe_eval_number(
                    self.too_ill_to_move[i])
                self.beta[i] = safe_eval_number(self.beta[i])
                self.contrib_foi[i] = safe_eval_number(self.contrib_foi[i])
            except Exception as e:
                raise AssertionError(
                    f"Invalid disease parameter at index {i}: "
                    f"{e.__class__} {e}")

        # for the model to work the different stages have set meanings
        errors = []

        # - stage 0 is newly infected that day, so progress must be 1
        #   and beta must be 0 (not infective)
        # This is not an error - the pox and flu2 diseases have this.
        # I don't think I understand correctly though if that is right,
        # as this means that individuals will stay longer in the
        # post-infect but pre-latent stage and be recorded as "recovered"?
        # if self.progress[0] != 1.0:
        #    errors.append(
        #        f"The progress[0] value must be 1.0 as individuals are "
        #        f"only newly infected for one day, and so must progress "
        #        f"immediately to the 'latent' stage.")

        if self.beta[0] != 0.0:
            errors.append(
                f"The beta[0] value must be 0.0 as newly infected "
                f"individuals should not be infective and cannot "
                f"infect others.")

        # - stage 1 is 'latent', meaning that beta must be 0 (not infective)
        if self.beta[1] != 0.0:
            errors.append(
                f"The beta[1] value must be 0.0 as 'latent' individuals "
                f"are not infectious and should not be able to infect "
                f"others.")

        # - stage -1 is 'recovered', meaning that beta must not be 0
        #   and progress is 0, as once recovered, always recovered
        if self.beta[-1] != 0.0:
            errors.append(
                f"The beta[-1] value must be 0.0 as 'recovered' individuals "
                f"are not infectious and should not be able to infect "
                f"others.")

        if self.progress[-1] != 0.0:
            errors.append(
                f"The progress[-1] value must be 0.0 as 'recovered' "
                f"individuals have no further disease stage to progress to. "
                f"We hope that once recovered, always recovered.")

        if len(errors) > 0:
            from .utils._console import Console
            Console.error("\n".join(errors))
            raise AssertionError("Invalid disease parameters!\n" +
                                 "\n".join(errors))

    @staticmethod
    def load(disease: str = "ncov",
             repository: str = None,
             folder: str = _default_folder_name,
             filename: str = None):
        """Load the disease parameters for the specified disease.
           This will look for a file called f"{disease}.json"
           in the directory f"{repository}/{folder}/{disease}.json"

           By default this will load the ncov (SARS-Cov-2)
           parameters from
           $HOME/GitHub/model_data/diseases/ncov.json

           Alternatively you can provide the full path to the
           json file via the "filename" argument

           Parameters
           ----------
           disease: str
             The name of the disease to load. This is the name that
             will be searched for in the METAWARDSDATA diseases directory
           repository: str
             The location of the cloned METAWARDSDATA repository
           folder: str
             The name of the folder within the METAWARDSDATA repository
             that contains the diseases
           filename: str
             The name of the file to load the disease from - this directly
             loads this file without searching through the METAWARDSDATA
             repository

           Returns
           -------
           disease: Disease
             The constructed and validated disease
        """
        repository_version = None
        repository_branch = None

        if filename is None:
            import os
            if os.path.exists(disease):
                filename = disease
            elif os.path.exists(f"{disease}.json"):
                filename = f"{disease}.json"

        if filename is None:
            from ._parameters import get_repository
            repository, v = get_repository(repository)

            filename = os.path.join(repository, folder,
                                    f"{disease}.json")

            repository = v["repository"]
            repository_version = v["version"]
            repository_branch = v["branch"]

        json_file = filename

        try:
            with open(json_file, "r") as FILE:
                import json
                data = json.load(FILE)

        except Exception as e:
            from .utils._console import Console
            Console.error(f"""
Could not find the disease file {json_file}. Either it does not exist of was
corrupted. Error was {e.__class__} {e}. Please see
https://metawards.org/model_data for instructions on how to download and
set the model data.""")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        disease = Disease(beta=data.get("beta", []),
                          progress=data.get("progress", []),
                          too_ill_to_move=data.get("too_ill_to_move", []),
                          contrib_foi=data.get("contrib_foi", []),
                          start_symptom=data.get("start_symptom", 3),
                          _name=disease,
                          _authors=data.get("author(s)", "unknown"),
                          _contacts=data.get("contact(s)", "unknown"),
                          _references=data.get("reference(s)", "none"),
                          _filename=json_file,
                          _repository=repository,
                          _repository_branch=repository_branch,
                          _repository_version=repository_version)

        disease._validate()

        return disease



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

from typing import List as _List

from libc.math cimport floor
from libc.stdint cimport uintptr_t

from .._nodes import Nodes
from .._links import Links
from .._network import Network
from ._profiler import Profiler

from ._get_array_ptr cimport get_double_array_ptr, get_int_array_ptr
from ._array import create_double_array

from ._ran_binomial cimport _ran_binomial, _ran_integer, \
                            _get_binomial_ptr, binomial_rng

__all__ = ["scale_node_susceptibles", "scale_link_susceptibles",
           "distribute_remainders"]


cdef inline double scale_and_round(double value, double scale) nogil:
    if scale > 0.5:
        # round up for large scales, as smaller scales will always
        # round down
        return floor((value * scale) + 0.5)
    else:
        # rounding down - hopefully this will minimise the number
        # of values that need to be redistributed
        return floor(value * scale)


def scale_node_susceptibles(nodes: Nodes, ratio: any = None,
                            work_ratio: any = None, play_ratio: any = None):
    """Scale the number of susceptibles in the passed Nodes
        by the passed scale ratios. These can be values, e.g.
        ratio = 2.0 will scale the total number of susceptibles
        in each ward by 2.0. They can also be lists of values,
        where ward[i] will be scaled by ratio[i]. They can also
        be dictionaries, e.g. ward[i] scaled by ratio[i]

        Parameters
        ----------
        nodes: Nodes
            The nodes whose susceptible population will be scaled
        ratio: None, float, list or dict
            The amount by which to scale the total population of
            susceptibles - evenly scales the work and play populations
        work_ratio: None, float, list or dict
            Scale only the work population of susceptibles
        play_ratio: None, float, list or dict
            Scale only the play population of susceptibles

        Returns
        -------
        None
    """

    if ratio is not None:
        work_ratio = ratio
        play_ratio = ratio

    if play_ratio is None:
        return

    if nodes is None or len(nodes) == 0:
        return

    try:
        play_ratio = float(play_ratio)
    except Exception:
        pass

    cdef double * nodes_play_suscept = get_double_array_ptr(
                                                nodes.play_suscept)
    cdef double * nodes_save_play_suscept = get_double_array_ptr(
                                                nodes.save_play_suscept)

    cdef int i = 0
    cdef int nnodes_plus_one = len(nodes)   # this will already
                                            # be the 1-indexed size

    cdef double v = 0.0
    cdef double val = 0.0

    if isinstance(play_ratio, float):
        if play_ratio == 1.0:
            return

        # scale all equally
        v = play_ratio

        for i in range(1, nnodes_plus_one):
            val = scale_and_round(nodes_save_play_suscept[i], v)
            nodes_play_suscept[i] = val
            nodes_save_play_suscept[i] = val

    elif isinstance(play_ratio, dict):
        # scale only the specified nodes
        for key, value in play_ratio.items():
            v = value
            val = scale_and_round(nodes_save_play_suscept[key], v)
            nodes_play_suscept[key] = val
            nodes_save_play_suscept[key] = val

    elif isinstance(play_ratio, list):
        if len(play_ratio) != len(nodes) - 1:
            raise ValueError(
                f"The list of play_ratio scale factors ({len(play_ratio)}) "
                f"must equal the number of nodes ({len(nodes)-1})")

        for i in range(1, nnodes_plus_one):
            v = play_ratio[i]
            val = scale_and_round(nodes_save_play_suscept[i], v)
            nodes_play_suscept[i] = val
            nodes_save_play_suscept[i] = val

    else:
        raise NotImplementedError(
                f"Cannot scale the nodes by a {play_ratio.__class__}")


def scale_link_susceptibles(links: Links, ratio: any):
    """Scale the number of susceptibles in the passed Links
        by the passed scale ratio. This can be a value, e.g.
        ratio = 2.0 will scale the total number of susceptibles
        by 2.0. This can also be lists of values,
        where ward[i] will be scaled by ratio[i]. They can also
        be dictionaries, e.g. ward[i] scaled by ratio[i]

        Parameters
        ----------
        ratio: None, float, list or dict
            The amount by which to scale the total population of
            susceptibles - evenly scales the work and play populations

        Returns
        -------
        None
    """

    if ratio is None:
        return

    if links is None or len(links) == 0:
        return

    try:
        ratio = float(ratio)
    except Exception:
        pass

    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)

    cdef int i = 0
    cdef int nlinks_plus_one = len(links)  # already 1-indexed size
    cdef int ifrom = 0
    cdef double v = 0.0
    cdef double val = 0

    if isinstance(ratio, float):
        if ratio == 1.0:
            return

        # scale all equally
        v = ratio

        for i in range(1, nlinks_plus_one):
            val = scale_and_round(links_weight[i], v)
            links_weight[i] = val
            links_suscept[i] = val

    elif isinstance(ratio, dict):
        # scale only links originating from specific nodes
        for i in range(1, nlinks_plus_one):
            ifrom = links_ifrom[i]

            v = ratio.get(ifrom, 1.0)

            if v != 1.0:
                val = scale_and_round(links_weight[i], v)
                links_weight[i] = val
                links_suscept[i] = val

    elif isinstance(ratio, list):
        # scale the links originating from the specified nodes
        # (the size of this list has already been checked in
        # the nodes equivalent of this function)
        for i in range(1, nlinks_plus_one):
            ifrom = links_ifrom[i]

            v = ratio[ifrom]

            if v != 1.0:
                val = scale_and_round(links_weight[i], v)
                links_weight[i] = val
                links_suscept[i] = val


cdef double redistribute(double target, double *values,
                         int nvalues, binomial_rng *rng) nogil:
    """This will add or subtract numbers from 'values' until their sum
       equals 'target'
    """
    cdef double diff = target

    cdef int i = 0

    for i in range(0, nvalues):
        diff = diff - values[i]

    if diff > 0:
        while diff > 0:
            i = _ran_integer(rng, 0, nvalues)
            values[i] += 1
            diff -= 1
    elif diff < 0:
        while diff < 0:
            i = _ran_integer(rng, 0, nvalues)
            values[i] -= 1
            diff += 1

    return diff


def distribute_remainders(network: Network,
                          subnets: _List[Network],
                          random_seed: int = None,
                          nthreads: int = 1,
                          profiler: Profiler=None) -> None:
    """Distribute the remainder of the population in each ward and link from
       'network' who are not represented in any of the demographic
       sub-networks in subnets. This uses a completely random algorithm
       that adds to random demographics one by one until all of the
       population is assigned

       Parameters
       ----------
       network: Network
         The overall network
       subnets: List[Network]
         The demographic sub-networks
       random_seed: int
         Random seed to use to seed the random number generator that
         decides where to allocate rounded individuals
    """

    from cython.parallel import parallel, prange

    if profiler is None:
        from ._profiler import NullProfiler
        profiler = NullProfiler()

    nodes = network.nodes
    links = network.links

    cdef int i = 0
    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef int nsubnets = len(subnets)

    cdef int n = 0
    cdef int sub_n = 0
    cdef int diff = 0

    cdef int num_threads = nthreads

    cdef double * nodes_play_suscept = get_double_array_ptr(
                                                nodes.play_suscept)
    cdef double * nodes_save_play_suscept = get_double_array_ptr(
                                                nodes.save_play_suscept)

    cdef double * sub_nodes_play_suscept
    cdef double * sub_nodes_save_play_suscept

    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)

    cdef double * sub_links_weight
    cdef double * sub_links_suscept

    diff_nodes = create_double_array(nnodes_plus_one)
    diff_links = create_double_array(nlinks_plus_one)

    cdef double * diff_nodes_array = get_double_array_ptr(diff_nodes)
    cdef double * diff_links_array = get_double_array_ptr(diff_links)

    cdef double target = 0.0
    values = create_double_array(nsubnets)
    cdef double * values_array = get_double_array_ptr(values)

    # calculate the number of remainders in each ward and link
    p = profiler.start("initialise")
    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nnodes_plus_one, schedule="static"):
            diff_nodes_array[i] = nodes_save_play_suscept[i]

        for i in prange(1, nlinks_plus_one, schedule="static"):
            diff_links_array[i] = links_weight[i]
    p = p.stop()

    p = p.start("calc_differences")
    for subnet in subnets:
        sub_nodes_save_play_suscept = get_double_array_ptr(
                                        subnet.nodes.save_play_suscept)
        sub_links_weight = get_double_array_ptr(subnet.links.weight)

        with nogil, parallel(num_threads=num_threads):
            for i in prange(1, nnodes_plus_one, schedule="static"):
                diff_nodes_array[i] = diff_nodes_array[i] - \
                                      sub_nodes_save_play_suscept[i]

            for i in prange(1, nlinks_plus_one, schedule="static"):
                diff_links_array[i] = diff_links_array[i] - \
                                      sub_links_weight[i]
    p = p.stop()

    # go through and take action on all of the differences. These differences
    # are rounding issues, e.g. dividing 5 people evenly between 2
    # demographics is not possible - we will assign 2 people to each
    # demographic, and would have 1 remainder. To solve this we could
    # either

    # 1. assign the person to the first demographic, or
    # 2. assign the person to the largest demographic, or
    # 3. assign the person to the smallest demographic, or
    # 4. assign the person to a random demographic

    # Solutions 1-3 have major issues that, because this happens for each
    # of the 1 million+ wards, we get massive biases to the first, or
    # largest, or smallest demographics (e.g. if this happens in all wards
    # then the "chosen" demographic has >1m more members than would be
    # expected). The solution appears to be that we assign randomly,
    # as, on average, the additions will occur to demographics evenly,
    # and so shouldn't affect their relative sizes.

    # HOWEVER, when we run multiple runs, we need the initial assignment
    # of population to demographics to be the same, otherwise this can
    # be a major confounding factor. What we need is for every time
    # MetaWards is run, that the same decision is taken about how to
    # round, and thus the same distribution of population amongst
    # demographics is each ward is made. The only way to achieve this
    # is to use a semi-random algorithm, i.e. we use random number
    # generator with a fixed random number seed to make the decisions.

    # While I don't like this, in most production use there will be
    # real data on the demographic splits, and so real, rather than
    # rounded values will be used and this algorithm will not be needed.

    from ._ran_binomial import seed_ran_binomial
    if random_seed is None:
        random_seed = 4751828

    from ._console import Console

    Console.print(
        f"Seeding generator used for demographics with seed {random_seed}")
    bin_rng = seed_ran_binomial(random_seed)

    cdef binomial_rng* rng = _get_binomial_ptr(bin_rng)

    p = p.start("distribute_nodes")
    with nogil:
        for i in range(1, nnodes_plus_one):
            if diff_nodes_array[i] != 0.0:
                target = nodes_save_play_suscept[i]

                for j in range(0, nsubnets):
                    with gil:
                        sub_nodes_save_play_suscept = get_double_array_ptr(
                                            subnets[j].nodes.save_play_suscept)

                    values_array[j] = sub_nodes_save_play_suscept[i]

                diff_nodes_array[i] = redistribute(target, values_array,
                                                   nsubnets, rng)

                for j in range(0, nsubnets):
                    with gil:
                        sub_nodes_save_play_suscept = get_double_array_ptr(
                                            subnets[j].nodes.save_play_suscept)
                        sub_nodes_play_suscept = get_double_array_ptr(
                                            subnets[j].nodes.play_suscept)

                    sub_nodes_save_play_suscept[i] = values_array[j]
                    sub_nodes_play_suscept[i] = values_array[j]
    p = p.stop()

    p = p.start("distribute_links")
    with nogil:
        for i in range(1, nlinks_plus_one):
            if diff_links_array[i] != 0.0:
                target = links_weight[i]

                for j in range(0, nsubnets):
                    with gil:
                        sub_links_weight = get_double_array_ptr(
                                                    subnets[j].links.weight)

                    values_array[j] = sub_links_weight[i]

                diff_links_array[i] = redistribute(target, values_array,
                                                   nsubnets, rng)

                for j in range(0, nsubnets):
                    with gil:
                        sub_links_weight = get_double_array_ptr(
                                                    subnets[j].links.weight)
                        sub_links_suscept = get_double_array_ptr(
                                                    subnets[j].links.suscept)

                    sub_links_weight[i] = values_array[j]
                    sub_links_suscept[i] = values_array[j]
    p = p.stop()

    Console.print(
        f"Number of differences is {sum(diff_nodes)} + {sum(diff_links)}")




from .._network import Network
from .._population import Population
from .._outputfiles import OutputFiles

from .._workspace import Workspace
from ..utils._get_functions import call_function_on_network

__all__ = ["output_incidence", "output_incidence_serial"]


def output_incidence_serial(network: Network,
                            population: Population,
                            output_dir: OutputFiles,
                            workspace: Workspace,
                            **kwargs):
    """This will incidence of infection for each ward for each timestep.
       This is the sum of infections from disease class 0 to 2 inclusive

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra argumentst that are ignored by this function
    """

    if network.name is None:
        name = ""
    else:
        name = "_" + network.name.replace(" ", "_")

    pfile = output_dir.open(f"incidence{name}.dat")

    pfile.write(str(population.day) + " ")

    pfile.write(" ".join([str(x) for x in workspace.incidence[1:]])
                + "\n")


def output_incidence(nthreads: int = 1, **kwargs):
    """This will incidence of infection for each ward for each timestep.
       This is the sum of infections from disease class 0 to 2 inclusive

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra argumentst that are ignored by this function
    """
    call_function_on_network(nthreads=1,
                             func=output_incidence_serial,
                             call_on_overall=True,
                             **kwargs)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
from libc.math cimport floor, ceil

from .._network import Network

from ._recalculate_denominators import recalculate_play_denominator_day, \
                                       recalculate_work_denominator_day
from ._profiler import Profiler
from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["move_population_from_work_to_play",
           "move_population_from_play_to_work"]


def move_population_from_work_to_play(network: Network,
                                      nthreads: int = 1,
                                      profiler: Profiler = None):
    """This function is not used or implemented, but is implied
       by the naming scheme...
    """
    raise AssertionError("Code has not been written")


def move_population_from_play_to_work(network: Network,
                                      nthreads: int = 1,
                                      profiler: Profiler = None):
    """And Vice Versa From Work to Play
       The relevant parameters are network.params.play_to_work and
       network.params.work_to_play

       When both are 0, don't do anything;
       When play_to_work > 0 move work_to_play proportion from play to work.
       When work_to_play > 0 move work_to_play proportion from work to play.
    """

    params = network.params

    if params is None:
        return

    links = network.links
    wards = network.nodes
    play = network.play

    cdef int i = 0
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef double * wards_play_suscept = get_double_array_ptr(wards.play_suscept)
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef double to_move = 0.0
    cdef double work_to_play = params.work_to_play
    cdef double play_to_work = params.play_to_work
    cdef double suscept = 0.0

    cdef int num_threads = nthreads
    cdef int nlinks_plus_one = network.nlinks + 1

    if params.work_to_play > 0.0:
        with nogil:
            for i in range(1, nlinks_plus_one):
                suscept = links_suscept[i]
                to_move = ceil(suscept * work_to_play)
                links_suscept[i] -= to_move
                wards_play_suscept[links_ifrom[i]] += to_move

    cdef int ifrom = 0
    cdef int * play_ifrom = get_int_array_ptr(play.ifrom)
    cdef double * play_weight = get_double_array_ptr(play.weight)
    cdef double * wards_save_play_suscept = get_double_array_ptr(
                                                    wards.save_play_suscept)
    cdef double countrem = 0.0
    cdef double temp = 0.0
    cdef double p = 0.0
    nlinks_plus_one = network.nplay + 1

    if params.play_to_work > 0.0:
        with nogil:
            for i in range(1, nlinks_plus_one):
                ifrom = play_ifrom[i]

                temp = play_to_work * (play_weight[i] *
                                    wards_save_play_suscept[ifrom])

                to_move = floor(temp)
                p = temp - to_move

                countrem += p

                if countrem >= 1.0:
                    to_move += 1.0
                    countrem -= 1.0

                if wards_play_suscept[ifrom] < to_move:
                    to_move = wards_play_suscept[ifrom]

                wards_play_suscept[ifrom] -= to_move
                links_suscept[i] += to_move

    recalculate_work_denominator_day(network=network, nthreads=nthreads,
                                     profiler=profiler)
    recalculate_play_denominator_day(network=network, nthreads=nthreads,
                                     profiler=profiler)




from array import array
import os

from .._network import Network
from .._infections import Infections

__all__ = ["allocate_vaccination",
           "how_many_vaccinated", "vaccinate_same_id"]


def allocate_vaccination(network: Network, output_dir: str):
    """Allocate memory and open files needed to track vaccination"""

    null_int = (network.nnodes+1) * [0]
    null_float = (network.nnodes+1) * [0.0]

    int_t = "i"
    float_t = "d"

    vac = array(int_t, null_int)
    wards_ra = array(int_t, null_int)
    risk_ra = array(float_t, null_float)
    sort_ra = array(int_t, null_int)
    VACF = open(os.path.join(output_dir, "Vaccinated.dat", "w"))

    trigger = 0

    return (vac, wards_ra, risk_ra, sort_ra, VACF, trigger)


def how_many_vaccinated(vac):
    raise AssertionError("how_many_vaccinated has not yet been written")


def vaccinate_same_id(network: Network, risk_ra, sort_ra,
                      infections: Infections,
                      vac, params):
    raise AssertionError("vaccinate_same_id has not yet been written")




from typing import List as _List
from typing import Dict as _Dict
from typing import Union as _Union

__all__ = ["VariableSets", "VariableSet"]


def _set_beta(params, name: str, index: int, value: float):
    """Adjust the Disease.beta parameter"""
    params.disease_params.beta[index] = float(value)


def _set_progress(params, name: str, index: int, value: float):
    """Adjust the Disease.progress parameter"""
    params.disease_params.progress[index] = float(value)


def _set_too_ill_to_move(params, name: str, index: int, value: float):
    """Adjust the Disease.too_ill_to_move parameter"""
    params.disease_params.too_ill_to_move[index] = float(value)


def _set_contrib_foi(params, name: str, index: int, value: float):
    """Adjust the Disease.contrib_foi parameter"""
    params.disease_params.contrib_foi[index] = float(value)


def _set_lengthday(params, name: str, index: int, value: float):
    """Adjust the Parameters.length_day parameter"""
    if index is not None:
        raise IndexError("You cannot index the lengthday")

    params.length_day = float(value)


def _set_plengthday(params, name: str, index: int, value: float):
    """Adjust the Parameters.plength_day parameter"""
    if index is not None:
        raise IndexError("You cannot index the lengthday")

    params.plength_day = float(value)


def _set_uv(params, name: str, index: int, value: float):
    """Adjust the Parameters.UV parameter"""
    if index is not None:
        raise IndexError("You cannot index the UV parameter")

    params.UV = float(value)


def _set_initial_inf(params, name: str, index: int, value: float):
    """Adjust the Parameters.initial_inf parameter"""
    if index is not None:
        raise IndexError("You cannot index the initial_inf parameter")

    params.initial_inf = float(value)


def _set_static_play_at_home(params, name: str, index: int, value: float):
    """Adjust the Parameters.static_play_at_home parameter"""
    if index is not None:
        raise IndexError("You cannot index the static_play_at_home parameter")

    params.static_play_at_home = float(value)


def _set_dyn_play_at_home(params, name: str, index: int, value: float):
    """Adjust the Parameters.dyn_play_at_home parameter"""
    if index is not None:
        raise IndexError("You cannot index the dyn_play_at_home parameter")

    params.dyn_play_at_home = float(value)


def _set_data_dist_cutoff(params, name: str, index: int, value: float):
    """Adjust the Parameters.data_dist_cutoff parameter"""
    if index is not None:
        raise IndexError("You cannot index the data_dist_cutoff parameter")

    params.data_dist_cutoff = float(value)


def _set_dyn_dist_cutoff(params, name: str, index: int, value: float):
    """Adjust the Parameters.dyn_dist_cutoff parameter"""
    if index is not None:
        raise IndexError("You cannot index the dyn_dist_cutoff parameter")

    params.dyn_dist_cutoff = float(value)


def _set_play_to_work(params, name: str, index: int, value: float):
    """Adjust the Parameters.play_to_work parameter"""
    if index is not None:
        raise IndexError("You cannot index the play_to_work parameter")

    params.play_to_work = float(value)


def _set_work_to_play(params, name: str, index: int, value: float):
    """Adjust the Parameters.work_to_play parameter"""
    if index is not None:
        raise IndexError("You cannot index the work_to_play parameter")

    params.work_to_play = float(value)


def _set_local_vaccination_thresh(params, name: str, index: int, value: float):
    """Adjust the Parameters.local_vaccination_thresh parameter"""
    if index is not None:
        raise IndexError("You cannot index the local_vaccination_thresh "
                         "parameter")

    params.local_vaccination_thresh = float(value)


def _set_global_detection_thresh(params, name: str, index: int, value: float):
    """Adjust the Parameters.global_detection_thresh parameter"""
    if index is not None:
        raise IndexError("You cannot index the global_detection_thresh "
                         "parameter")

    params.global_detection_thresh = float(value)


def _set_daily_ward_vaccination_capacity(params, name: str,
                                         index: int, value: float):
    """Adjust the Parameters.daily_ward_vaccination_capacity
       parameter
    """
    if index is not None:
        raise IndexError("You cannot index the daily_ward_vaccination "
                         "parameter")

    params.daily_ward_vaccination_capacity = float(value)


def _set_neighbour_weight_threshold(params, name: str,
                                    index: int, value: float):
    """Adjust the Parameters.neighbour_weight_threshold parameter"""
    if index is not None:
        raise IndexError("You cannot index the neighbour_weight_threshold "
                         "parameter")

    params.neighbour_weight_threshold = float(value)


def _set_daily_imports(params, name: str, index: int, value: float):
    """Adjust the Parameters.daily_imports parameter"""
    if index is not None:
        raise IndexError("You cannot index the daily_imports parameter")

    params.daily_imports = float(value)


def _set_user_params(params, name: str, index: int, value: float):
    """Adjust a custom user-supplied parameter, held in
       Parameters.user_params[name]. Set a user parameter
       called 'name' via 'user.name' or '.name'.
    """
    if name.startswith("user."):
        name = name[5:]
    elif name.startswith("."):
        name = name[1:]

    if params.user_params is None:
        params.user_params = {}

    if index is None:
        params.user_params[name] = value
    else:
        if name not in params.user_params:
            params.user_params[name] = []

        while len(params.user_params[name]) <= index:
            params.user_params[name].append(None)

        params.user_params[name][index] = value


_adjustable = {}
_adjustable["beta"] = _set_beta
_adjustable["progress"] = _set_progress
_adjustable["too_ill_to_move"] = _set_too_ill_to_move
_adjustable["contrib_foi"] = _set_contrib_foi
_adjustable["user"] = _set_user_params
_adjustable["length_day"] = _set_lengthday
_adjustable["plength_day"] = _set_plengthday
_adjustable["UV"] = _set_uv
_adjustable["initial_inf"] = _set_initial_inf
_adjustable["static_play_at_home"] = _set_static_play_at_home
_adjustable["dyn_play_at_home"] = _set_dyn_play_at_home
_adjustable["data_dist_cutoff"] = _set_data_dist_cutoff
_adjustable["dyn_dist_cutoff"] = _set_dyn_dist_cutoff
_adjustable["play_to_work"] = _set_play_to_work
_adjustable["work_to_play"] = _set_work_to_play
_adjustable["local_vaccination_thesh"] = _set_local_vaccination_thresh
_adjustable["global_detection_thresh"] = _set_global_detection_thresh
_adjustable["daily_ward_vaccination_capacity"] = \
    _set_daily_ward_vaccination_capacity
_adjustable["neighbour_weight_threshold"] = _set_neighbour_weight_threshold
_adjustable["daily_imports"] = _set_daily_imports


def _clean(x):
    """Clean the passed string by stripping off unnecesary characters,
       and turning "True" and "False" into 1 and 0. Also change
       '=' and ':' into '=='
    """
    x = x.strip()

    if x.lower() == "true":
        return True
    elif x.lower() == "false":
        return False
    elif x == "=" or x == ":":
        return "=="
    elif x.endswith("=="):
        return x[0:-2], "=="
    elif x.endswith(":") or x.endswith("="):
        return x[0:-1], "=="
    else:
        return x


def _interpret(value):
    if not isinstance(value, str):
        return value

    canonical = value.lower().replace('"', "'")

    if canonical.startswith("d'"):
        # this is a date
        try:
            from dateparser import parse
            return parse(value[2: -1]).date()
        except Exception:
            pass

        from datetime import date
        return date.fromisoformat(value[2: -1])

    elif canonical.startswith("f'"):
        # this is a floating point number
        return float(value[2: -1])

    elif canonical.startswith("i'"):
        # this is an integer
        return int(value[2:-1])

    elif canonical.startswith("s'"):
        # this is a string
        return value[2:-1]

    # now we have to guess...
    try:
        v = float(value)
        if v.is_integer():
            return int(v)
        else:
            return v
    except Exception:
        pass

    try:
        from datetime import date
        return date.fromisoformat(value)
    except Exception:
        pass

    try:
        from .utils._safe_eval import safe_eval_number
        return safe_eval_number(value)
    except Exception:
        pass

    # do this last as it is quite slow...
    try:
        from dateparser import parse
        return parse(value).date()
    except Exception:
        pass

    # this can only be a string...
    return value


def _wrap(text, width=70):
    """Return 'text' wrapped to at most 'width' characters, as
       a list of lines
    """
    import textwrap
    text = text.strip().replace(r"\s+", " ").replace("\n", " ")
    return textwrap.wrap(text, width)


class VariableSet:
    """This class holds a single set of adjustable variables that
       are used to adjust the variables as part of a model run

       Examples
       --------
       >>> v = VariableSet()
       >>> v["beta[1]"] = 0.95
       >>> v["beta[2]"] = 0.9
       >>> print(v.fingerprint())
       (beta[1]=0.95, beta[2]=0.9)[repeat 1]
       >>> params = Parameters()
       >>> params.set_disease("ncov")
       >>> v.adjust(params)
       >>> print(params.disease_params.beta[1],
       >>>       params.disease_params.beta[2])
       0.95 0.9
    """

    def __init__(self,
                 variables: _Dict[str, float] = None,
                 repeat_index: int = 1,
                 names: _List[str] = None,
                 values: _List[float] = None):
        """Construct a new VariableSet from the passed adjusted variable
           values.

           Parameters
           ----------
           names: List[str]
             The list of the names of the variables to adjust
           values: List[float]
             The values of the variables to adjust (same order as names)
           variables: Dict[str, float]
             names and values of variables to adjust passed as a dictionary
           repeat_index: int
             the index used to distinguish different repeats of the same
             VariableSet from one another

           Examples
           --------
           >>> v = VariableSet()
           >>> v["beta[1]"] = 0.95
           >>> v["beta[2]"] = 0.9
           >>> print(v.fingerprint())
           (beta[1]=0.95, beta[2]=0.9)[repeat 1]
        """
        self._names = None
        self._vals = None
        self._varnames = None
        self._varidxs = None
        self._idx = None
        self._output = None

        if variables is not None:
            for name, value in variables.items():
                self._add(name, value)

        if values is not None:
            if names is None:
                names = ["beta[2]", "beta[3]", "progress[1]",
                         "progress[2]", "progress[3]"]

            if len(names) != len(values):
                raise IndexError(
                    f"The number of variable values '{values}' must equal "
                    f"the number of variable names '{names}'")

            for name, value in zip(names, values):
                self._add(name, value)

        self._idx = repeat_index

    def __str__(self):
        """Return a printable representation of the variables to
           be adjusted
        """
        s = []

        if self._vals is not None and len(self._vals) > 0:
            for key, value in zip(self._names, self._vals):
                s.append(f"{key}={value}")

        if len(s) == 0:
            return f"(NO_CHANGE)[repeat {self._idx}]"
        else:
            return f"({', '.join(s)})[repeat {self._idx}]"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, dict):
            other = VariableSet(variables=other)

        if self._idx != other._idx:
            return False

        if len(self) != len(other):
            return False

        if self._vals is None:
            return

        for i in range(0, len(self._vals)):
            if self._vals[i] != other._vals[i]:
                return False

            if self._varnames[i] != other._varnames[i]:
                return False

            if self._varidxs[i] != other._varidxs[i]:
                return False

            if self._names[i] != other._names[i]:
                return False

        return True

    def __len__(self):
        if self._vals is None:
            return 0
        else:
            return len(self._vals)

    def __getitem__(self, key):
        if self._vals is None:
            raise KeyError(f"No adjustable parameter {key} in an empty set")

        if key.startswith("user."):
            key = key[4:]

        for i, name in enumerate(self._names):
            if key == name:
                return self._vals[i]

        raise KeyError(f"No adjustable parameter {key}. Available parameters "
                       f"are '{self._names}'")

    def __setitem__(self, key, value):
        if self._names is None:
            self._names = []

        for i, name in enumerate(self._names):
            if key == name:
                self._vals[i] = value
                return

        self._add(key, value)

    def _add(self, name, value):
        """Internal function to add a new variable called 'name' to
           be varied - it will be set equal to 'value'
        """
        import re

        if self._vals is None:
            self._names = []
            self._vals = []
            self._varnames = []
            self._varidxs = []

        name = name.strip()

        if name == "output":
            # this is the output directory for this run. This will
            # replace the default name, which is based on the fingerprint
            value = str(value).strip()

            if len(value) == 0:
                raise ValueError("You cannot use an empty string as the "
                                 "output directory")

            self._output = value
            return

        # look for 'variable[index]'
        m = re.search(r"([\.\w]+)\[(\d+)\]", name)

        if m:
            varname = m.group(1)
            index = int(m.group(2))
            value = value
        else:
            varname = name
            index = None
            value = value

        if not (varname.startswith("user.") or varname.startswith(".") or
                varname in _adjustable):
            raise KeyError(f"It is not possible to adjust the variable "
                           f"{name} to equal {value}. Available variables "
                           f"are {list(_adjustable.keys())}, or to set "
                           f"a user parameter 'user.parameter' or "
                           f"'.parameter'. To set an index use "
                           f"'parameter[index]', e.g. 'beta[2]'")

        if varname.startswith("user."):
            # strip off 'user' so that it just starts with a dot
            varname = varname[4:]

        if name.startswith("user."):
            name = name[4:]

        self._varnames.append(varname)
        self._varidxs.append(index)
        self._names.append(name)
        self._vals.append(value)

    @staticmethod
    def adjustable_help():
        """Return a string that contains some help that is useful
           for finding out which variables can be adjusted
        """
        lines = []

        names = list(_adjustable.keys())
        names.sort()

        for name in names:
            func = _adjustable[name]
            docs = _wrap(func.__doc__, 60)
            lines.append("")
            lines.append(f"{name}:")

            for doc in docs:
                lines.append(f"  {doc}")

        return "\n".join(lines)

    def variable_names(self):
        """Return the names of the variables that will be adjusted
           by this VariableSet

           Returns
           -------
           names: List[str]
             The list of names of variables to be adjusted
        """
        if self._vals is None or len(self._vals) == 0:
            return None
        else:
            from copy import deepcopy
            return deepcopy(self._names)

    def variable_values(self):
        """Return the values that the variables will be adjusted to.
           Note that 'None' means that the variable won't be adjusted
           from its default (original) value

           Returns
           -------
           values: List[float]
             The list of values for variables to be adjusted to
        """
        if self._vals is None or len(self._vals) == 0:
            return None
        else:
            from copy import deepcopy
            return deepcopy(self._vals)

    def variables(self):
        """Return the variables (name and values) to be adjusted

           Returns
           -------
           variables: Dict[str, float]
             The dictionary mapping the names of the variables that
             with be adjusted to their desired values
        """
        v = {}

        for name, value in zip(self._names, self._vals):
            v[name] = value

        return v

    def repeat_index(self):
        """Return the repeat index of this set. The repeat index is the
           ID of this set if the VariableSet is repeated. The index should
           range from 1 to nrepeats

           Returns
           -------
           index: int
             The repeat index of this set
        """
        return self._idx

    def make_compatible_with(self, other):
        """Return a copy of this VariableSet which has been made
           compatible with 'other'. This means that it will change
           the same variables as 'other', e.g. by adding 'None'
           changes for missing variables. This will raise an error
           if it is not possible to make this set compatible

           Parameters
           ----------
           other: VariableSet
             The passed VariableSet for which this should be made compatible

           Returns
           -------
           result: VariableSet
             A copy of this VariableSet which is now compatible with 'other'

           Example
           -------
           >>> v1 = VariableSet()
           >>> v1["beta[1]"] = 0.9
           >>> v1["beta[2]"] = 0.8

           >>> v2 = VariableSet()
           >>> v2["beta[1]"] = 0.6
           >>> v2 = v2.make_compatible_with(v1)
           >>> print(v2)
           (beta[1]=0.6, beta[2]=0.8)[repeat 1]
        """
        from copy import deepcopy

        if self._names is None:
            v = deepcopy(other)
            v._idx = self._idx
            return v

        if other._names is None:
            raise ValueError(f"VariableSet {self} is not compatible with "
                             f"VariableSet {other}")

        nmatch = 0

        for name in self._names:
            if name not in other._names:
                raise ValueError(f"VariableSet {self} is not compatible with "
                                 f"VariableSet {other}")
            nmatch += 1

        if len(other._names) == nmatch:
            # fully compatible
            return deepcopy(self)

        v = deepcopy(self)

        for name in other._names:
            if name not in self._names:
                v[name] = other[name]

        return v

    @staticmethod
    def _extract_values(fingerprint: str):
        fingerprint = str(fingerprint)

        # has this fingerprint been added onto the end of a filename
        # using a hyphen or underscore? If so, then remove this
        if fingerprint.find("-") != -1:
            fingerprint = fingerprint.split("-")[-1]

        if fingerprint.find("_") != -1:
            fingerprint = fingerprint.split("_")[-1]

        fingerprint.strip()

        # does this have a repeat index - if so, this is the last
        # value at the end after the 'x'
        parts = fingerprint.split("x")

        if len(parts) > 1:
            repeat_index = int(parts[-1])
            fingerprint = "x".join(parts[0:-1])
        else:
            repeat_index = None

        # now get the values
        values = []

        if fingerprint != "REPEAT":
            for part in fingerprint.split("v"):
                try:
                    if part == "T":
                        values.append(True)
                    elif part == "F":
                        values.append(False)
                    else:
                        values.append(float(part.replace("i", ".")))
                except Exception:
                    # this is not part of the fingerprint
                    pass

        return (values, repeat_index)

    @staticmethod
    def extract_values(fingerprint: str):
        """Return the original values from the passed fingerprint
           or filename. This assumes that the fingerprint
           was created using the 'fingerprint' function, namely
           that any integers are actually 0.INTEGER

           Parameters
           ----------
           fingerprint: str
             The fingerprint (or filename) to decode

           Returns
           -------
           (values, repeat): (List[float], int)
             The list of values of the variables and the repeat
             index. The repeat index is None if it wasn't included
             in the fingerprint
        """
        from pathlib import Path

        path = fingerprint

        # if this is a filename then extract only the fingerprint
        fingerprint = Path(Path(fingerprint).name)
        for suffix in fingerprint.suffixes:
            fingerprint = str(fingerprint).replace(suffix, "")

        values, repeat_idx = VariableSet._extract_values(fingerprint)

        if len(values) > 0 or repeat_idx is not None:
            return (values, repeat_idx)

        import os

        if path.find(os.path.sep) != -1:
            fingerprint = os.path.sep.join(
                path.split(os.path.sep)[0:-1])
            return VariableSet.extract_values(fingerprint)
        else:
            return (values, repeat_idx)

    @staticmethod
    def create_fingerprint(vals: _List[float], index: int = None,
                           include_index: bool = False):
        """Create the fingerprint for the passed values"""
        f = None

        if vals is None or len(vals) == 0:
            f = "REPEAT"
        else:
            for val in vals:
                if isinstance(val, bool):
                    if val:
                        v = "T"
                    else:
                        v = "F"
                else:
                    v = float(val)

                    if v.is_integer():
                        v = int(val)
                        v = f"{v}.0"

                    v = str(v).replace(".", "i")

                if f is None:
                    f = v
                else:
                    f += "v" + v

        if include_index:
            return "%sx%03d" % (f, index)
        else:
            return f

    def fingerprint(self, include_index: bool = False):
        """Return a fingerprint for this VariableSet. This can be
           used to quickly identify and distinguish the values of
           the variables in this set from the values in other
           VariableSets which have the same adjustable variables,
           but different parameters

           Parameters
           ----------
           include_index: bool
             Whether or not to include the repeat_index in the fingerprint

           Returns
           -------
           fingerprint: str
             The fingerprint for this VariableSet
        """
        return VariableSet.create_fingerprint(vals=self._vals,
                                              index=self._idx,
                                              include_index=include_index)

    def output_dir(self):
        """Return the output directory in which runs using this
           variable set should be placed. Normally this would
           be the fingerprint of the variable set, but users may
           prefer to specify their own naming scheme, which
           can be added via a design file
        """
        if self._output is None:
            return self.fingerprint(include_index=True)
        else:
            return self._output

    @staticmethod
    def read(filename: str):
        """Read a single set of adjustable variables from the passed
           file. The file can either write the variables in horizontal
           or vertical mode, using space or comma separated values.

           This is useful for when you want to set a global set of parameters
           at the start of a calculation and don't want to use
           a large VariableSets

           Parameters
           ----------
           filename: str
             The name of the file containing the VariableSet

           Returns
           -------
           variables: VariableSet
             The VariableSet that has been read
        """
        variables = VariableSets.read(filename)

        if len(variables) == 0:
            return VariableSet()
        elif len(variables) != 1:
            raise ValueError(
                "More than one set of variables was read from {filename}")
        else:
            return variables[0]

    def adjust(self, params):  # should be 'Parameters' but circular include
        """Use the variables in this set to adjust the passed parameters.
           Note that this directly modifies 'params'

           Parameters
           ----------
           params: Parameters
             The parameters whose variables will be adjusted

           Returns
           -------
           None

           Examples
           --------
           >>> v = VariableSet()
           >>> v["beta[1]"] = 0.95
           >>> v["beta[2]"] = 0.9
           >>> print(v.fingerprint())
           (beta[1]=0.95, beta[2]=0.9)[repeat 1]
           >>> params = Parameters()
           >>> params.set_disease("ncov")
           >>> v.adjust(params)
           >>> print(params.disease_params.beta[1],
           >>>       params.disease_params.beta[2])
           0.95 0.9
        """
        if self._vals is None or len(self._vals) == 0:
            return

        try:
            for varname, varidx, value in zip(self._varnames, self._varidxs,
                                              self._vals):
                if varname.startswith("user.") or varname.startswith("."):
                    _adjustable["user"](params=params,
                                        name=varname,
                                        index=varidx,
                                        value=value)
                elif varname in _adjustable:
                    _adjustable[varname](params=params,
                                         name=varname,
                                         index=varidx,
                                         value=value)
                else:
                    raise KeyError(
                        f"Cannot set unrecognised parameter {varname} "
                        f"to {value}")
        except Exception as e:
            raise ValueError(
                f"Unable to set parameters from {self}. Error "
                f"equals {e.__class__}: {e}")


class VariableSets:
    """This class holds the collection of all VariableSet objects
       that contain the set of adjustable variables that are used
       to control a single run of the model

       Examples
       --------
       >>> v = VariableSets()
       >>> v.append({"beta[2]": 0.95, "beta[3]": 0.9})
       >>> v.append({"beta[1]": 0.86, "beta[2]": 0.89})
       >>> print(v)
       {(beta[2]=0.95, beta[3]=0.9)[repeat 1], (beta[1]=0.86,
       beta[2]=0.89)[repeat 1]}
       >>> v = v.repeat(2)
       >>> print(v)
       {(beta[2]=0.95, beta[3]=0.9)[repeat 1], (beta[1]=0.86,
       beta[2]=0.89)[repeat 1], (beta[2]=0.95, beta[3]=0.9)[repeat 2],
       (beta[1]=0.86, beta[2]=0.89)[repeat 2]}
    """

    def __init__(self):
        """Initialise an empty VariableSets object

           Parameters
           ----------
           None

           Returns
           -------
           None
        """
        self._vars = []

    def __str__(self):
        s = []

        for v in self._vars:
            s.append(str(v))

        if len(s) == 1:
            return s[0]
        elif len(s) > 0:
            return "{" + ",\n ".join(s) + "\n}"
        else:
            return "VariableSets:empty"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, dict):
            v = VariableSet(variables=other)
            other = VariableSets()
            other.append(v)

        if len(self._vars) != len(other._vars):
            return False

        for v0, v1 in zip(self._vars, other._vars):
            if v0 != v1:
                return False

        return True

    def __len__(self):
        if self._vars is None:
            return 0
        else:
            return len(self._vars)

    def __getitem__(self, i: int):
        """Return the VariableSet at the specified index"""
        if self._vars is None:
            raise IndexError("Cannot index an empty VariableSets object")
        else:
            return self._vars[i]

    def append(self, variables: VariableSet):
        """Append the passed set of variables to the set that will
           be used to run a model. If there are any existing
           VariableSet objects in this list, then the new VariableSet
           must adjust the same variables

           Parameters
           ----------
           variables: VariableSet
             The VariableSet to append to this list. If you pass a
             dict of {str: float} values, then this will automatically
             be converted into a VariableSet. Note that all VariableSet
             objects in a VariableSets must adjust the same variables

           Returns
           -------
           None
        """
        if isinstance(variables, dict):
            variables = VariableSet(variables=variables)

        if self._vars is None:
            self._vars = []

        if len(self._vars) > 0:
            variables = variables.make_compatible_with(self._vars[0])

        self._vars.append(variables)

    def repeat(self, nrepeats: _Union[_List[int], int]):
        """Return a copy of this VariableSet in which all of the
           unique VaribleSet objects have been repeated 'nrepeats'
           times

           Parameters
           ----------
           nrepeats: int or list[int]
             The number of repeats of the VariableSet objects to
             perform. If this is a list, then nrepeats[i] will be
             the number of times to repeat variables[i]

           Returns
           -------
           repeats: VariableSets
             A new VariableSets object containing 'nrepeats' copies
             of the VariableSet objects from this set
        """
        if not isinstance(nrepeats, list):
            nrepeats = [nrepeats]

        if len(nrepeats) == 1 and nrepeats[0] <= 1:
            return self

        from copy import deepcopy

        repeats = VariableSets()

        if len(nrepeats) == 1:
            nrepeats = nrepeats[0]
            for i in range(1, nrepeats+1):
                for v in self._vars:
                    v2 = deepcopy(v)
                    v2._idx = i
                    repeats.append(v2)
        else:
            if len(nrepeats) != len(self._vars):
                raise ValueError(
                    f"Disagreement of the number of repeats {len(nrepeats)} "
                    f"and the number of variables {len(self._vars)}")

            added = True
            n = 1
            while added:
                added = False

                for i, v in enumerate(self._vars):
                    if n <= nrepeats[i]:
                        v2 = deepcopy(v)
                        v2._idx = n
                        repeats.append(v2)
                        added = True

                n += 1

        return repeats

    @staticmethod
    def read(filename: str, line_numbers: _List[int] = None):
        """Read and return collection of VariableSet objects from the
           specified line number(s) of the specified file

           Parameters
           ----------
           filename: str
             The name of the file from which to read the VariableSets
           line_numbers: List[int]
             The line numbers from the file to read. This is 0-indexed,
             meaning that the first line is line 0. If this is None,
             then all lines are read and used

           Returns
           -------
           variables: VariableSets
             The collection of VariableSet objects that have been read,
             in the order they were read from the file
        """
        if not isinstance(line_numbers, list):
            if line_numbers is not None:
                line_numbers = [line_numbers]

        # parse all lines using the csv module
        import csv
        lines = open(filename, "r").readlines()

        csvlines = []
        for line in lines:
            line = line.strip()

            if len(line) > 0 and (not line.startswith("#")):
                csvlines.append(line)

        lines = []

        # first try to guess the dialect of the file (space or comma
        # separated, newline character etc.)
        try:
            dialect = csv.Sniffer().sniff(csvlines[0], delimiters=[" ", ","])
        except Exception:
            from .utils._console import Console
            Console.warning(
                f"Could not identify what sort of separator to use to "
                f"read {filename}, so will assume commas. If this is wrong, "
                f"then could you add commas to separate the "
                f"fields?")
            dialect = csv.excel  #  default comma-separated file

        for line in csv.reader(csvlines, dialect=dialect,
                               quoting=csv.QUOTE_ALL,
                               skipinitialspace=True):
            cleaned = [_clean(x) for x in line]

            line = []

            if len(cleaned) > 0:
                for clean in cleaned:
                    if isinstance(clean, list) or isinstance(clean, tuple):
                        for c in clean:
                            if len(c) > 0:
                                line.append(c)
                    else:
                        line.append(clean)

                lines.append(line)

        if len(lines) == 0:
            # there is nothing to read?
            return VariableSets()

        if len(lines[0]) > 1 and len(lines[0]) > 1 and lines[0][1] == "==":
            # this is a vertical file
            if line_numbers is not None:
                raise ValueError(
                    "You cannot specify line numbers for a vertical file!")
            return VariableSets._read_vertical(lines)
        else:
            return VariableSets._read_horizontal(lines=lines,
                                                 line_numbers=line_numbers)

    @staticmethod
    def _read_vertical(lines):
        """Read the data from the vertical lines"""
        variables = VariableSets()
        variable = VariableSet()

        for line in lines:
            if len(line) > 1:
                key = line[0]

                if line[1] == "==":
                    if len(line) > 3:
                        value = " ".join(line[2:])
                    else:
                        value = line[2]
                else:
                    if len(line) > 2:
                        value = " ".join(line[1:])
                    else:
                        value = line[1]

                variable[key] = _interpret(value)

        variables.append(variable)

        return variables

    @staticmethod
    def _read_horizontal(lines, line_numbers):
        """Read the data from the horizontal lines"""
        variables = VariableSets()

        # are there any strings on the first line? If so, then these
        # are the titles
        has_titles = False
        for v in lines[0]:
            try:
                float(v)
            except Exception:
                has_titles = True
                break

        if has_titles:
            titles = lines[0]
            lines = lines[1:]
        else:
            # default adjustable variables
            titles = ["beta[2]", "beta[3]", "progress[1]",
                      "progress[2]", "progress[3]"]

        if "repeats" in titles:
            repeats_index = titles.index("repeats")
            titles.pop(repeats_index)
        else:
            repeats_index = None

        repeats = []

        for i, line in enumerate(lines):
            if line_numbers is None or i in line_numbers:
                values = [_interpret(x) for x in line]

                if repeats_index is not None:
                    repeats.append(int(values.pop(repeats_index)))

                variable = VariableSet()

                for j, key in enumerate(titles):
                    variable[key] = values[j]

                variables.append(variable)

                if line_numbers is not None:
                    if len(line_numbers) == len(variables):
                        if repeats_index is not None:
                            variables = variables.repeat(repeats)

                        return variables

        # get here if we can't find this line in the file (or if we
        # are supposed to read all lines)
        if line_numbers is None:
            if repeats_index is not None:
                variables = variables.repeat(repeats)
            return variables
        else:
            raise ValueError(
                f"Cannot read parameters from line {line_numbers} "
                f"as the number of lines in the file is {i+1}")



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

from libc.stdio cimport FILE, fopen, fscanf, fclose, feof

from .._parameters import Parameters
from .._network import Network
from .._nodes import Nodes
from .._links import Links

from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

from ._profiler import Profiler, NullProfiler

__all__ = ["build_wards_network"]


def _read_network(filename: str, max_nodes: int, max_links: int):
    """This function reads in the network of nodes and links
       from the passed file, returning a tuple of (Nodes, Links)
    """
    nodes = Nodes(max_nodes + 1)     # need to pre-allocate nodes and links
    links = Links(max_links + 1)   # both of these use 1-indexing

    cdef int MAX_LINKS = max_links
    cdef int MAX_NODES = max_nodes

    cdef int nlinks = 0
    cdef int nnodes = 0

    cdef int from_id = 0
    cdef int to_id = 0
    cdef double weight = 0.0
    cdef int iweight = 0

    cdef int * nodes_begin_to = get_int_array_ptr(nodes.begin_to)
    cdef int * nodes_end_to = get_int_array_ptr(nodes.end_to)
    cdef int * nodes_self_w = get_int_array_ptr(nodes.self_w)
    cdef int * nodes_label = get_int_array_ptr(nodes.label)

    cdef int * links_ito = get_int_array_ptr(links.ito)
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)

    cdef double * nodes_denominator_d = get_double_array_ptr(
                                                    nodes.denominator_d)
    cdef double * nodes_denominator_n = get_double_array_ptr(
                                                    nodes.denominator_n)

    # need to use C file reading as too slow in python
    f = filename.encode("UTF-8")
    cdef char* fname = f

    cdef FILE* cfile
    cfile = fopen(fname, "r")

    cdef int error_from = -1
    cdef int error_to = -1

    if cfile == NULL:
        raise FileNotFoundError(f"No such file or directory: {filename}")

    with nogil:
        while not feof(cfile):
            fscanf(cfile, "%d %d %lf\n", &from_id, &to_id, &weight)

            iweight = <int>weight

            if from_id == 0 or to_id == 0:
                error_from = from_id
                error_to = to_id
                break

            nlinks += 1

            if nlinks >= MAX_LINKS:
                break

            if from_id > nnodes:
                nnodes = from_id

            if to_id > nnodes:
                nnodes = to_id

            if nnodes >= MAX_NODES:
                break

            if nodes_label[from_id] == -1:
                nodes_label[from_id] = from_id
                nodes_begin_to[from_id] = nlinks
                nodes_end_to[from_id] = nlinks

            if from_id == to_id:
                nodes_self_w[from_id] = nlinks

            nodes_end_to[from_id] += 1

            links_ifrom[nlinks] = from_id
            links_ito[nlinks] = to_id
            links_weight[nlinks] = weight
            links_suscept[nlinks] = weight

            nodes_denominator_n[from_id] += weight
            nodes_denominator_d[to_id] += weight

        fclose(cfile)

    if nlinks >= MAX_LINKS or nnodes >= MAX_NODES:
        raise MemoryError(
            f"There are either too many links (>{nlinks}) or too many "
            f"wards (>{nnodes}) to fit into pre-allocated memory "
            f"(max_nodes = {MAX_NODES}, max_links = {MAX_LINKS}). "
            f"Increase these values and try to run again.")

    if error_from != -1 or error_to != -1:
        raise ValueError(f"{filename} is corrupted! "
                         f"Zero in link list: ${error_from}-${error_to}! "
                         f"Renumber files and start again")

    return (nodes, links, nnodes, nlinks)


_data_cache = {}


def build_wards_network(params: Parameters,
                        profiler: Profiler = None,
                        nthreads: int = 1,
                        max_nodes:int = 16384,
                        max_links:int = 4194304):
    """Creates a network of wards using the information provided in
       the file specified in parameters.input_files.work.

       The format of this should be:

        * Node_1 Node_2 weight 1-2
        * Node_3 Node_4 weight 3-4
        * Node_4 Node_1 weight 4-1
        * Node_2 Node_1 weight 2-1
        * ...

         BE CAREFUL!! Weights may not be symmetric, network is built with
         asymmetric links
    """
    global _data_cache

    # the network is solely determined by the work and play files
    work_and_play = f"{params.input_files.work}:{params.input_files.play}"

    from ._console import Console

    if work_and_play in _data_cache:
        Console.print("Using pre-loaded cached network...")
        network = _data_cache[work_and_play].copy()
        network.params = params
        return network

    if profiler is None:
        profiler = NullProfiler()

    p = profiler.start("build_wards_network")

    workfile = params.input_files.work


    p = p.start("read_work_file")
    if workfile in _data_cache:
        Console.print("Using pre-loaded cached work matrix...")
        (nodes, links, nnodes, nlinks) = _data_cache[workfile]
        nodes = nodes.copy()
        links = links.copy()

    else:
        try:
            (nodes, links,
             nnodes, nlinks) = _read_network(filename=workfile,
                                             max_nodes=max_nodes,
                                             max_links=max_links)
        except MemoryError as e:
            Console.print(f"Increasing max_nodes to {max_nodes*2} and "
                          f"max_links to {max_links*2}")
            return build_wards_network(params=params,
                                       profiler=profiler,
                                       nthreads=nthreads,
                                       max_nodes=max_nodes*2,
                                       max_links=max_links*2)

        _data_cache[workfile] = (nodes.copy(), links.copy(), nnodes, nlinks)
    p = p.stop()

    network = Network(nnodes=nnodes, nlinks=nlinks)

    network.nodes = nodes
    network.links = links

    Console.print(f"Number of nodes equals {nnodes}")
    Console.print(f"Number of links equals {nlinks}")

    # save the parameters used to build the network
    # within the network - this will save having to pass
    # them separately, which is error-prone
    network.params = params
    network.max_links = max_links
    network.max_nodes = max_nodes

    from . import fill_in_gaps
    p = p.start("fill_in_gaps")
    fill_in_gaps(network, max_nodes=max_nodes)
    p = p.stop()

    Console.print(f"Number of nodes after filling equals {network.nnodes}")

    from . import build_play_matrix
    p = p.start("build_play_matrix")
    build_play_matrix(network=network, profiler=p, max_nodes=max_nodes,
                      max_links=max_links)
    p = p.stop()

    # now finally go through all of the nodes and make sure that their
    # label is not -1
    cdef int nnull = 0
    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int * nodes_label = get_int_array_ptr(nodes.label)
    cdef int i

    with nogil:
        for i in range(1, nnodes_plus_one):
            if nodes_label[i] == -1:
                nnull += 1
                nodes_label[i] = i

    if nnull > 0:
        Console.print(f"Number of null nodes equals {nnull}")

    Console.print(f"Number of nodes after build play equals {network.nnodes}")

    Console.print(f"Resize nodes to {network.nnodes + 1}")
    Console.print(f"Resize links to {network.nlinks + 1}")
    Console.print(f"Resize play links to {network.nplay + 1}")

    p = p.start("resize_nodes_and_links")
    network.nodes.resize(network.nnodes + 1)     # remember 1-indexed
    network.links.resize(network.nlinks + 1)  # remember 1-indexed
    network.play.resize(network.nplay + 1)      # remember 1-indexed
    p = p.stop()

    # cache this network for future use
    _data_cache[work_and_play] = network.copy()

    p.stop()

    return network




from typing import List as _List

__all__ = ["InteractionMatrix"]


class InteractionMatrix:
    """This is an interaction matrix, which is used to control
       how the FOIs of different demographics are merged together.

       An interaction matrix is just a square matrix, with
       a list[][] being perfectly acceptable. This is really
       a convenience class that makes it easier to create
       more complex interaction matrixes
    """

    def __init__(self, n: int, value: float = 0.0):
        """Construct an interaction matrix that is n x n in size,
           where all values equal 'value'
        """
        n = int(n)

        if n <= 0:
            raise ValueError(
                "You cannot create a zero-sized interaction matrix")

        value = float(value)

        self._matrix = []

        for i in range(0, n):
            row = [value] * n
            self._matrix.append(row)

    def __getitem__(self, i: int):
        """Return the ith row of the matrix"""
        return self._matrix[i]

    def __len__(self):
        return len(self._matrix)

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        for i, row in enumerate(self._matrix):
            if len(row) != len(other[i]):
                return False

            for j, value in enumerate(row):
                if value != other[i][j]:
                    return False

        return True

    def __setitem__(self, i: int, row: _List[int]):
        """"Set the ith row of the matrix equal to 'row'"""
        if len(row) != len(self):
            raise ValueError(f"Incorrect row size {len(row)} for a "
                             f"{len(self)} by {len(self)} interaction matrix")

    @staticmethod
    def ones(n: int, value: float = 1.0):
        """Return a n x n matrix where each element equals 'value'"""
        return InteractionMatrix(n=n, value=value)

    @staticmethod
    def zeroes(n: int, value: float = 0.0):
        """Return a n x n matrix where each element equals 'value'"""
        return InteractionMatrix.ones(n=n, value=value)

    @staticmethod
    def diagonal(n: int, value: float = 1.0,
                 off_diagonal: float = 0.0):
        """Return a n x n matrix where each diagonal element equals
           'value' and each off-diagonal element equals 'off_diagonal'
        """
        m = InteractionMatrix(n=n, value=off_diagonal)

        for i in range(0, n):
            m[i][i] = value

        return m

    @staticmethod
    def identity(n: int, value: float = 1.0,
                 off_diagonal: float = 0.0):
        """Return a n x n matrix where each diagonal element equals
           'value' and each off-diagonal element equals 'off_diagonal'
        """
        return InteractionMatrix.diagonal(n=n, value=value,
                                          off_diagonal=off_diagonal)

    def __str__(self):
        lines = []

        for row in self._matrix:
            r = ["%5.3f" % x for x in row]
            lines.append("| " + ", ".join(r) + " |")

        return "\n".join(lines)

    def resize(self, n: int, value: float = 0.0):
        """Resize this matrix to 'n x n', adding in extra
           values equal to 'value' if needed
        """
        if len(self) == n:
            return

        elif n < len(self):
            m = InteractionMatrix(n=n)

            for i in range(0, n):
                for j in range(0, n):
                    m[i][j] = self[i][j]

            self._matrix = m._matrix

        else:
            m = InteractionMatrix(n=n, value=value)

            n = len(self)

            for i in range(0, n):
                for j in range(0, n):
                    m[i][j] = self[i][j]

            self._matrix = m._matrix

    def detach(self, n: int):
        """Detach the 'nth' demographic from interacting with
           any other demographics. This sets the ith row and ith
           column equal to zero (which not changing m[n][n])
        """
        for i in range(len(self)):
            if i != n:
                self[i][n] = 0.0
                self[n][i] = 0.0




from .._network import Network

__all__ = ["add_lookup"]


def add_lookup(network: Network, nthreads: int = 1):
    """Add in metadata about the network that can be used
       to look up wards by name of location or region etc.

       This will add the data to the network.ward_info object,
       as a list ensuring that network.nodes[i] has its
       info in network.ward_info[i]
    """
    lookup_file = network.params.input_files.lookup

    if lookup_file is None:
        from ._console import Console
        Console.print("No ward lookup information available.")
        return

    lines = open(lookup_file, "r").readlines()

    if lines[0].find(",") != -1:
        sep = ","
    else:
        sep = " "

    columns = network.params.input_files.lookup_columns

    WDCD = columns.get("code", None)
    WDNM = columns.get("name", None)
    CMWDCD = columns.get("alternate_code", None)
    CMWDNM = columns.get("alternate_name", None)
    LADCD = columns.get("authority_code", None)
    LADNM = columns.get("authority_name", None)
    REGCD = columns.get("region_code", None)
    REGNM = columns.get("region_name", None)

    from .._wardinfo import WardInfo, WardInfos

    ward_infos = []
    ward_infos.append(None)   # 1-indexed

    import csv

    for parts in csv.reader(lines[1:], quotechar='"', delimiter=',',
                            quoting=csv.QUOTE_ALL, skipinitialspace=True):

        info = WardInfo()

        if WDCD is not None:
            info.code = parts[WDCD].strip()

        if WDNM is not None:
            info.name = parts[WDNM].strip()

        if CMWDCD is not None:
            info.alternate_codes.append(parts[CMWDCD].strip())

        if CMWDNM is not None:
            info.alternate_names.append(parts[CMWDNM].strip())

        if LADCD is not None:
            info.authority_code = parts[LADCD].strip()

        if LADNM is not None:
            info.authority = parts[LADNM].strip()

        if REGCD is not None:
            info.region_code = parts[REGCD].strip()

        if REGNM is not None:
            info.region = parts[REGNM].strip()

        ward_infos.append(info)

    if len(ward_infos) != network.nnodes + 1:
        from ._console import Console
        Console.warning(
            f"Number of wards from {lookup_file} "
            f"({len(ward_infos)}) disagrees with the number "
            f"of wards in the network ({network.nnodes})")

        if len(ward_infos) > network.nnodes+1:
            ward_infos = ward_infos[0:network.nnodes+1]
        else:
            while len(ward_infos) <= network.nnodes:
                ward_infos.append(None)

    network.info = WardInfos(wards=ward_infos)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
cimport openmp

from libc.math cimport floor

from .._network import Network

from ._profiler import Profiler
from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["recalculate_work_denominator_day",
           "recalculate_play_denominator_day"]


def recalculate_work_denominator_day(network: Network, nthreads: int = 1,
                                     profiler: Profiler = None):
    """Recalculate the denominator_d for the wards (nodes) in
       the network for the normal links
    """
    params = network.params

    if params is None:
        return

    wards = network.nodes
    links = network.links

    cdef int i = 0

    cdef double * wards_denominator_d = get_double_array_ptr(
                                                wards.denominator_d)
    cdef double * wards_denominator_n = get_double_array_ptr(
                                                wards.denominator_n)

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int num_threads = nthreads

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nnodes_plus_one, schedule="static"):
            wards_denominator_d[i] = 0.0
            wards_denominator_n[i] = 0.0

    cdef int j = 0
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef int ifrom = 0
    cdef int ito = 0
    cdef int nlinks_plus_one = network.nlinks + 1
    cdef double suscept = 0.0
    cdef double sum = 0

    # not easily parallelisable due to reduction
    for j in range(1, nlinks_plus_one):
        ifrom = links_ifrom[j]
        ito = links_ito[j]
        suscept = links_suscept[j]
        wards_denominator_d[ito] += suscept
        wards_denominator_n[ifrom] += suscept
        sum += suscept

    #print(f"recalculate_work_denominator_day sum = {sum}")


def recalculate_play_denominator_day(network: Network, nthreads: int = 1,
                                     profiler: Profiler = None):
    """Recalculate the denominator_d for the wards (nodes) in
       the network for the play links
    """
    params = network.params

    if params is None:
        return

    wards = network.nodes
    links = network.play

    cdef int i = 0
    cdef double * wards_denominator_pd = get_double_array_ptr(
                                                    wards.denominator_pd)
    cdef double * wards_denominator_p = get_double_array_ptr(
                                                    wards.denominator_p)

    cdef int num_threads = nthreads
    cdef int nnodes_plus_one = network.nnodes + 1

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nnodes_plus_one, schedule="static"):
            wards_denominator_pd[i] = 0
            wards_denominator_p[i] = 0

    cdef double sum = 0.0
    cdef int j = 0
    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)
    cdef int ifrom = 0
    cdef int ito = 0
    cdef double weight = 0.0
    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double denom = 0.0
    cdef double * wards_play_suscept = get_double_array_ptr(wards.play_suscept)

    cdef int nlinks_plus_one = network.nplay + 1

    # not easily parallelisable due to the reduction
    with nogil:
        for j in range(1, nlinks_plus_one):
            ifrom = links_ifrom[j]
            ito = links_ito[j]
            weight = links_weight[j]
            denom = weight * wards_play_suscept[ifrom]
            wards_denominator_pd[ito] += denom

            sum += denom

    #print(f"recalculate_play_denominator_day sum 1 = {sum}")

    sum = 0.0
    cdef double play_suscept = 0
    cdef double pd = 0.0

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nnodes_plus_one, schedule="static"):
            pd = wards_denominator_pd[i]
            play_suscept = wards_play_suscept[i]

            wards_denominator_pd[i] = floor(pd + 0.5)
            wards_denominator_p[i] = play_suscept

            sum += play_suscept

    #print(f"recalculate_play_denominator_day sum 2 = {sum}")



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange

from .._networks import Networks

from ..utils._profiler import Profiler
from ..utils._get_array_ptr cimport get_double_array_ptr
from ..utils._array import create_double_array

__all__ = ["merge_using_matrix"]


def merge_using_matrix(network: Networks, nthreads: int,
                       profiler: Profiler, **kwargs):
    """This merge_function merges the FOIs across all demographic
       sub-networks according to the interaction matrix stored
       in networks.demographics.interaction_matrix.
    """

    matrix = network.demographics.interaction_matrix

    if matrix is None:
        # No matrix, so nothing should interact
        return

    subnets = network.subnets
    cdef int nsubnets = len(subnets)

    if nsubnets < 2:
        # nothing to merge
        return

    # the matrix should be sized for the number of subnets
    if len(matrix) != nsubnets:
        raise ValueError(
            f"The interaction matrix must be right-sized for the number "
            f"of demographics, e.g. it must be {nsubnets}x{nsubnets}")

    # it must also be square
    for row in matrix:
        if len(row) != nsubnets:
            raise ValueError(
                f"The interaction matrix must be square, e.g. "
                f"{nsubnets}x{nsubnets}.")

    # if all values are 1.0 then it is quicker to call merge_evenly

    # similarly if this is [[1,0], [0,1]] then it is quicker to
    # call merge_none. However, we won't do this now as this is a good
    # test if this merge function is correct

    cdef int nnodes_plus_one = network.overall.nnodes + 1

    cdef int i = 0
    cdef int j = 0
    cdef int k = 0

    cdef int num_threads = nthreads

    wards = network.overall.nodes

    cdef double scl = 0.0

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * day_foi
    cdef double * night_foi

    cdef double * sub_day_foi
    cdef double * sub_night_foi

    # create space to calculate all of the day and night fois
    day_fois = []
    night_fois = []

    p = profiler.start("allocate")
    for i in range(0, nsubnets):
        day_fois.append(create_double_array(nnodes_plus_one, 0.0))
        night_fois.append(create_double_array(nnodes_plus_one, 0.0))
    p = p.stop()

    # calculate the merged FOIs
    p = profiler.start("accumulate")
    for i in range(0, nsubnets):
        day_foi = get_double_array_ptr(day_fois[i])
        night_foi = get_double_array_ptr(night_fois[i])

        for j in range(0, nsubnets):
            sub_wards = subnets[j].nodes
            sub_day_foi = get_double_array_ptr(sub_wards.day_foi)
            sub_night_foi = get_double_array_ptr(sub_wards.night_foi)

            scl = matrix[i][j]

            with nogil, parallel(num_threads=num_threads):
                for k in prange(1, nnodes_plus_one, schedule="static"):
                    day_foi[k] = day_foi[k] + \
                                 scl * sub_day_foi[k]
                    night_foi[k] = night_foi[k] + \
                                   scl * sub_night_foi[k]
    p = p.stop()

    p = profiler.start("distribute")
    for j in range(0, nsubnets):
        day_foi = get_double_array_ptr(day_fois[j])
        night_foi = get_double_array_ptr(night_fois[j])

        sub_wards = subnets[j].nodes
        sub_day_foi = get_double_array_ptr(sub_wards.day_foi)
        sub_night_foi = get_double_array_ptr(sub_wards.night_foi)

        with nogil, parallel(num_threads=num_threads):
            for k in prange(1, nnodes_plus_one, schedule="static"):
                sub_day_foi[k] = day_foi[k]
                sub_night_foi[k] = night_foi[k]
    p = p.stop()




__all__ = ["extract_small"]


def extract_small(**kwargs):
    """This extractor only extracts the 'small' default files,
       e.g. nothing that involves the incidence or prevalence
       matrices
    """

    from ._output_basic import output_basic
    from ._output_dispersal import output_dispersal

    return [output_basic, output_dispersal]



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
from libc.stdint cimport uintptr_t

from .._network import Network
from .._infections import Infections

from ..utils._profiler import Profiler
from ..utils._get_functions import call_function_on_network

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["advance_play", "advance_play_omp",
           "advance_play_serial"]


def advance_play_omp(network: Network, infections: Infections, rngs,
                     nthreads: int, profiler: Profiler, **kwargs):
    """Advance the model by triggering infections related to random
       'play' movements (parallel version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections: Infections
         The space that holds all of the infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    play = network.play
    params = network.params

    play_infections = infections.play

    # Copy arguments from Python into C cdef variables
    cdef int * wards_begin_p = get_int_array_ptr(wards.begin_p)
    cdef int * wards_end_p = get_int_array_ptr(wards.end_p)

    cdef int * play_ito = get_int_array_ptr(play.ito)
    cdef double * play_weight = get_double_array_ptr(play.weight)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * wards_play_suscept = get_double_array_ptr(wards.play_suscept)
    cdef double * play_distance = get_double_array_ptr(play.distance)

    cdef double * wards_day_inf_prob = get_double_array_ptr(
                                                    wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                    wards.night_inf_prob)

    cdef double dyn_play_at_home = params.dyn_play_at_home
    cdef double cutoff = params.dyn_dist_cutoff

    # Pointer to the play_infections array - only need [0] as this loop
    # is creating new infections
    cdef int * play_infections_i = get_int_array_ptr(play_infections[0])

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    # create and initialise variables used in the loop
    cdef int num_threads = nthreads
    cdef int thread_id = 0

    cdef int j = 0
    cdef int k = 0
    cdef int l = 0
    cdef int ito = 0
    cdef int nnodes_plus_one = network.nnodes + 1

    cdef double weight = 0.0
    cdef double inf_prob = 0.0
    cdef double prob_scaled = 0.0
    cdef double cumulative_prob = 0.0

    cdef int suscept = 0
    cdef int staying = 0
    cdef int moving = 0

    cdef int play_move = 0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all wards to create
    ## new infections that appear in those wards at
    ## daytime and nighttime

    p = profiler.start("play")
    with nogil, parallel(num_threads=num_threads):
        thread_id = cython.parallel.threadid()
        rng = _get_binomial_ptr(rngs_view[thread_id])

        for j in prange(1, nnodes_plus_one, schedule="static"):
            inf_prob = 0.0
            suscept = <int>wards_play_suscept[j]
            staying = _ran_binomial(rng, dyn_play_at_home, suscept)

            moving = suscept - staying

            cumulative_prob = 0.0

            # daytime infection of play matrix moves
            for k in range(wards_begin_p[j], wards_end_p[j]):
                if play_distance[k] < cutoff:
                    ito = play_ito[k]

                    if wards_day_foi[ito] > 0.0:
                        weight = play_weight[k]
                        prob_scaled = weight / (1.0-cumulative_prob)
                        cumulative_prob = cumulative_prob + weight

                        play_move = _ran_binomial(rng, prob_scaled, moving)
                        inf_prob = wards_day_inf_prob[ito]

                        l = _ran_binomial(rng, inf_prob, play_move)

                        moving = moving - play_move

                        if l > 0:
                            # infection
                            play_infections_i[j] += l
                            wards_play_suscept[j] -= l
                    # end of DayFOI if statement
                # end of Dynamics Distance if statement
            # end of loop over links of wards[j]

            if (staying + moving) > 0:
                # infect people staying at home
                inf_prob = wards_day_inf_prob[j]
                l = _ran_binomial(rng, inf_prob, staying+moving)

                if l > 0:
                    # another infections, this time from home
                    #print(f"staying home play_infections[{i}][{j}] += {l}")
                    play_infections_i[j] += l
                    wards_play_suscept[j] -= l

            # nighttime infections of play movements
            inf_prob = wards_night_inf_prob[j]
            if inf_prob > 0.0:
                l = _ran_binomial(rng, inf_prob, <int>(wards_play_suscept[j]))

                if l > 0:
                    # another infection
                    play_infections_i[j] += l
                    wards_play_suscept[j] -= l
        # end of loop over wards (nodes)
    # end of parallel
    p.stop()


def advance_play_serial(network: Network, infections: Infections, rngs,
                        profiler: Profiler, **kwargs):
    """Advance the model by triggering infections related to random
       'play' movements (serial version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections: Infections
         The space that holds all of the infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    links = network.links
    wards = network.nodes
    play = network.play
    params = network.params

    play_infections = infections.play

    # Copy arguments from Python into C cdef variables
    cdef int * wards_begin_p = get_int_array_ptr(wards.begin_p)
    cdef int * wards_end_p = get_int_array_ptr(wards.end_p)

    cdef int * play_ito = get_int_array_ptr(play.ito)
    cdef double * play_weight = get_double_array_ptr(play.weight)

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * wards_play_suscept = get_double_array_ptr(wards.play_suscept)
    cdef double * play_distance = get_double_array_ptr(play.distance)

    cdef double * wards_day_inf_prob = get_double_array_ptr(
                                                    wards.day_inf_prob)
    cdef double * wards_night_inf_prob = get_double_array_ptr(
                                                    wards.night_inf_prob)

    cdef double dyn_play_at_home = params.dyn_play_at_home
    cdef double cutoff = params.dyn_dist_cutoff

    # Pointer to the play_infections array - only need [0] as this loop
    # is creating new infections
    cdef int * play_infections_i = get_int_array_ptr(play_infections[0])

    # get the random number generator
    cdef binomial_rng* rng = _get_binomial_ptr(rngs[0])

    cdef int j = 0
    cdef int k = 0
    cdef int l = 0
    cdef int ito = 0
    cdef int nnodes_plus_one = network.nnodes + 1

    cdef double weight = 0.0
    cdef double inf_prob = 0.0
    cdef double prob_scaled = 0.0
    cdef double cumulative_prob = 0.0

    cdef int suscept = 0
    cdef int staying = 0
    cdef int moving = 0

    cdef int play_move = 0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all wards to create
    ## new infections that appear in those wards at
    ## daytime and nighttime

    p = profiler.start("play")
    with nogil:
        for j in range(1, nnodes_plus_one):
            inf_prob = 0.0
            suscept = <int>wards_play_suscept[j]
            staying = _ran_binomial(rng, dyn_play_at_home, suscept)

            moving = suscept - staying

            cumulative_prob = 0.0

            # daytime infection of play matrix moves
            for k in range(wards_begin_p[j], wards_end_p[j]):
                if play_distance[k] < cutoff:
                    ito = play_ito[k]

                    if wards_day_foi[ito] > 0.0:
                        weight = play_weight[k]
                        prob_scaled = weight / (1.0-cumulative_prob)
                        cumulative_prob = cumulative_prob + weight

                        play_move = _ran_binomial(rng, prob_scaled, moving)
                        inf_prob = wards_day_inf_prob[ito]

                        l = _ran_binomial(rng, inf_prob, play_move)

                        moving = moving - play_move

                        if l > 0:
                            # infection
                            play_infections_i[j] += l
                            wards_play_suscept[j] -= l
                    # end of DayFOI if statement
                # end of Dynamics Distance if statement
            # end of loop over links of wards[j]

            if (staying + moving) > 0:
                # infect people staying at home
                inf_prob = wards_day_inf_prob[j]
                l = _ran_binomial(rng, inf_prob, staying+moving)

                if l > 0:
                    # another infections, this time from home
                    #print(f"staying home play_infections[{i}][{j}] += {l}")
                    play_infections_i[j] += l
                    wards_play_suscept[j] -= l

            # nighttime infections of play movements
            inf_prob = wards_night_inf_prob[j]
            if inf_prob > 0.0:
                l = _ran_binomial(rng, inf_prob, <int>(wards_play_suscept[j]))

                if l > 0:
                    # another infection
                    play_infections_i[j] += l
                    wards_play_suscept[j] -= l
        # end of loop over wards (nodes)
    # end of nogil
    p.stop()

def advance_play(nthreads: int, **kwargs):
    """Advance the model by triggering infections related to random
       'play' movements (parallel version of the function)

       Parameters
       ----------
       network: Network
         The network being modelled
       infections: Infections
         The space that holds all of the "play" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """
    call_function_on_network(nthreads=nthreads,
                             func=advance_play_serial,
                             parallel=advance_play_omp,
                             **kwargs)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange

from .._networks import Networks

from ..utils._profiler import Profiler
from ..utils._get_array_ptr cimport get_double_array_ptr

__all__ = ["merge_evenly"]


def merge_evenly(network: Networks, nthreads: int,
                 profiler: Profiler, **kwargs):
    """This merge_function merges the FOIs across all demographic
       sub-networks evenly. This should be equivalent to
       calculating the FOIs of the entire network without
       using demographics
    """

    subnets = network.subnets
    cdef int nsubnets = len(subnets)

    if nsubnets < 2:
        # nothing to merge
        return

    cdef int nnodes_plus_one = network.overall.nnodes + 1

    cdef int i = 0
    cdef int j = 0

    cdef int num_threads = nthreads

    wards = network.overall.nodes

    cdef double * wards_day_foi = get_double_array_ptr(wards.day_foi)
    cdef double * wards_night_foi = get_double_array_ptr(wards.night_foi)

    cdef double * sub_day_foi
    cdef double * sub_night_foi

    # accumulate the FOI for each ward into the overall network
    p = profiler.start("accumulate")
    with nogil, parallel(num_threads=num_threads):
        for j in prange(1, nnodes_plus_one, schedule="static"):
            wards_day_foi[j] = 0.0
            wards_night_foi[j] = 0.0

    for subnet in subnets:
        sub_wards = subnet.nodes
        sub_day_foi = get_double_array_ptr(sub_wards.day_foi)
        sub_night_foi = get_double_array_ptr(sub_wards.night_foi)

        with nogil, parallel(num_threads=num_threads):
            for j in prange(1, nnodes_plus_one, schedule="static"):
                wards_day_foi[j] = wards_day_foi[j] + \
                                   sub_day_foi[j]
                wards_night_foi[j] = wards_night_foi[j] + \
                                     sub_night_foi[j]

    p = p.stop()

    # now copy this total FOI back to all demographics
    p = p.start("distribute")

    for subnet in subnets:
        sub_wards = subnet.nodes
        sub_day_foi = get_double_array_ptr(sub_wards.day_foi)
        sub_night_foi = get_double_array_ptr(sub_wards.night_foi)

        with nogil, parallel(num_threads=num_threads):
            for j in prange(1, nnodes_plus_one, schedule="static"):
                sub_day_foi[j] = wards_day_foi[j]
                sub_night_foi[j] = wards_night_foi[j]

    p = p.stop()




from dataclasses import dataclass as _dataclass
from dataclasses import field as _field
from typing import List as _List
from typing import Dict as _Dict

from metawards.utils import Console

__all__ = ["SpringFlowers"]


@_dataclass
class SpringFlowers:
    """This is the colourful 'SpringFlowers' theme"""

    frames: _Dict[int, _List[str]] = _field(default_factory=dict)

    panel_colors: _List[str] = _field(default_factory=list)

    panel_color_count = 0

    def should_highlight(self):
        return False

    def highlighter(self):
        return None

    def should_markup(self):
        return False

    def text(self, style):
        if style == "warning":
            return "magenta"
        elif style == "error":
            return "red"
        elif style == "info":
            return "cyan"
        else:
            return "white"

    def error(self):
        return "red"

    def warning(self):
        return "magenta"

    def info(self):
        return "cyan"

    def spinner_success(self, spinner):
        if Console.supports_emojis():
            spinner.green.ok("✔")
        else:
            spinner.green.ok("Success")

    def spinner_failure(self, spinner):
        if Console.supports_emojis():
            spinner.red.fail("✘")
        else:
            spinner.red.fail("Failed")

    def rule(self, style):
        if style is None:
            return "green"
        elif style == "finish":
            return "magenta"
        elif style == "error":
            return self.error()
        elif style == "warning":
            return self.warning()
        elif style == "info":
            return self.info()
        elif style == "iteration":
            return "cyan"
        else:
            return "cyan"

    def panel_box(self, style):
        from rich import box as _box

        if style == "header":
            return _box.HEAVY_EDGE
        elif style == "command":
            return _box.MINIMAL_HEAVY_HEAD
        else:
            return _box.SQUARE

    def padding_style(self, style):
        if style == "header":
            return "on #220077"
        elif style == "command":
            return "bold white on #222222"
        else:
            return self.panel(style, advance=False)

    def panel(self, style, advance=True):
        if style is None:
            return "on black"

        elif style == "command":
            return "white on #222222"

        elif style == "alternate":
            if len(self.panel_colors) == 0:
                self.panel_colors = ["blue", "cyan"]
                self.panel_color_count = 0

            color = self.panel_colors[self.panel_color_count]

            if advance:
                self.panel_color_count += 1
                if self.panel_color_count >= len(self.panel_colors):
                    self.panel_color_count = 0

            return f"on {color}"

        elif style == "header":
            return f"on #0000FF"

        else:
            return "on black"

    def get_frames(self, width: int = 80):
        """Return the frames used to animate a spinner in a console
           of specified width

           This returns the list of frames plus the timeout between
           the list
        """
        if width in self.frames:
            return self.frames[width]

        frames = []
        frames.append("")

        if Console.supports_emojis():
            bar = "👉 👉 👉 😷  😷  😷 👌 👍 👏 👏 👏 👏 👏 "
        else:
            bar = "-> -> -> #WearAMask :-) :-) :-)  "

        for i in range(1, len(bar), 1):
            frames.append(bar[0:i])

        self.frames[width] = (frames, 50)

        return self.frames[width]




from dataclasses import dataclass

__all__ = ["Link"]


@dataclass
class Link:
    """This is a link between nodes in the network"""

    #: Index of the origin node (ward)
    ifrom: int = None

    #: Index of the destination node (ward)
    ito: int = None

    #: Weight of the link, used to save the original number
    #: of susceptibles in the work matrix
    weight: float = None

    #: Number of susceptibles in the case of the work matrix. Used
    #: to save the weight of the play matrix
    suscept: float = None

    #: The distance between the two wards connected by this link
    distance: float = None




from .._network import Network
from .._population import Population
from .._outputfiles import OutputFiles
from .._workspace import Workspace

from ..utils._get_functions import call_function_on_network

__all__ = ["output_wards_trajectory", "output_wards_trajectory_serial"]


def output_wards_trajectory_serial(network: Network,
                                   population: Population,
                                   output_dir: OutputFiles,
                                   workspace: Workspace,
                                   **kwargs):
    """This will output the complete trajectory for
       S, E, I and R for each of the wards in the model.
       This may take a lot of disk space (10's MBs).

       The files are written to name "wards_trajectory_X.dat",
       where "X" is one of S, E, I or R

       Parameters
       ----------
       population: Population
         Model population - used to get the day
       output_dir: OutputFiles
         Where to place the output files
       workspace: Workspace
         Workspace containing the raw data
       **kwargs:
         Other arguments not needed by this function
    """

    if network.name is None:
        name = ""
    else:
        name = "_" + network.name.replace(" ", "_")

    S_file = output_dir.open(f"wards_trajectory{name}_S.dat")
    E_file = output_dir.open(f"wards_trajectory{name}_E.dat")
    I_file = output_dir.open(f"wards_trajectory{name}_I.dat")
    R_file = output_dir.open(f"wards_trajectory{name}_R.dat")

    day = str(population.day) + " "

    S_file.write(day)
    E_file.write(day)
    I_file.write(day)
    R_file.write(day)

    S_file.write(" ".join([str(x) for x in workspace.S_in_wards]))
    E_file.write(" ".join([str(x) for x in workspace.E_in_wards]))
    I_file.write(" ".join([str(x) for x in workspace.I_in_wards]))
    R_file.write(" ".join([str(x) for x in workspace.R_in_wards]))

    S_file.write("\n")
    E_file.write("\n")
    I_file.write("\n")
    R_file.write("\n")


def output_wards_trajectory(nthreads: int = 1, **kwargs):
    """This will output the complete trajectory for
       S, E, I and R for each of the wards in the model.
       This may take a lot of disk space (10's MBs).

       The files are written to name "wards_trajectory_X.dat",
       where "X" is one of S, E, I or R

       Parameters
       ----------
       population: Population
         Model population - used to get the day
       output_dir: OutputFiles
         Where to place the output files
       workspace: Workspace
         Workspace containing the raw data
       **kwargs:
         Other arguments not needed by this function
    """
    call_function_on_network(nthreads=1,
                             func=output_wards_trajectory_serial,
                             call_on_overall=True,
                             **kwargs)




__all__ = ["import_module"]


def import_module(module):
    """This will try to import the passed module. This will return
       the module if it was imported, or will return 'None' if
       it should not be imported.

       Parameters
       ----------
       module: str
         The name of the module to import
    """

    from ._console import Console

    try:
        import importlib
        m = importlib.import_module(module)
    except SyntaxError as e:
        Console.error(
            f"\nSyntax error when importing {module}\n"
            f"{e.__class__.__name__}:{e}\n"
            f"Line {e.lineno}.{e.offset}:{(e.offset-1)*' '} |\n"
            f"Line {e.lineno}.{e.offset}:{(e.offset-1)*' '}\\|/\n"
            f"Line {e.lineno}.{e.offset}: {e.text}")
        m = None
    except ImportError:
        # this is ok and expected if the module is in a python file
        # that will be loaded below
        m = None
    except Exception:
        Console.error(
            f"\nError when importing {module}\n"
            f"{e.__class__.__name__}:{e}\n")
        m = None

    if m is None:
        try:
            import importlib.util
            import os

            if os.path.exists(module):
                pyfile = module
            elif os.path.exists(f"{module}.py"):
                pyfile = f"{module}.py"
            elif os.path.exists(f"{module}.pyx"):
                pyfile = f"{module}.pyx"
            else:
                pyfile = None

            if pyfile:
                spec = importlib.util.spec_from_file_location(
                    module,
                    pyfile)

                m = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(m)

                Console.print(f"Loaded {module} from {pyfile}")

        except SyntaxError as e:
            Console.error(
                f"\nSyntax error when reading {pyfile}\n"
                f"{e.__class__.__name__}:{e}\n"
                f"Line {e.lineno}.{e.offset}:{(e.offset-1)*' '} |\n"
                f"Line {e.lineno}.{e.offset}:{(e.offset-1)*' '}\\|/\n"
                f"Line {e.lineno}.{e.offset}: {e.text}")
        except Exception as e:
            Console.error(
                f"\nError when importing {module}\n"
                f"{e.__class__.__name__}:{e}\n")
            m = None

    Console.print(f"IMPORT {m}")

    return m



"""
.. currentmodule:: metawards.iterators

Functions
=========

.. autosummary::
    :toctree: generated/

    advance_additional
    advance_additional_omp
    advance_fixed
    advance_fixed_omp
    advance_foi
    advance_foi_omp
    advance_imports
    advance_imports_omp
    advance_infprob
    advance_infprob_omp
    advance_play
    advance_play_omp
    advance_recovery
    advance_recovery_omp
    build_custom_iterator
    iterate_custom
    iterate_default
    iterate_weekday
    iterate_weekend
    iterate_working_week
    setup_additional_seeds
    setup_seed_all_wards
    setup_seed_specified_ward
    setup_seed_wards
"""

from ._advance_additional import *
from ._advance_fixed import *
from ._advance_foi import *
from ._advance_imports import *
from ._advance_infprob import *
from ._advance_play import *
from ._advance_recovery import *

from ._iterate_custom import *
from ._iterate_default import *
from ._iterate_weekday import *
from ._iterate_weekend import *
from ._iterate_working_week import *

from ._setup_imports import *




from dataclasses import dataclass as _dataclass
from dataclasses import field as _field
from typing import List as _List

__all__ = ["WardInfo", "WardInfos"]


@_dataclass
class WardInfo:
    """This class holds metadata about a ward, e.g. its name(s),
       any ID code(s), any information about the region or
       authority it is in etc.
    """
    #: Name of the ward
    name: str = ""
    #: Any alternative names of the ward
    alternate_names: _List[str] = _field(default_factory=list)

    #: Official ID code of the ward
    code: str = ""
    #: Any alternative ID codes of the ward
    alternate_codes: _List[str] = _field(default_factory=list)

    #: The name of the local authority it is in
    authority: str = ""

    #: The ID of the local authority it is in
    authority_code: str = ""

    #: The name of the region it is in
    region: str = ""

    #: The ID of the region it is in
    region_code: str = ""


@_dataclass
class WardInfos:
    """Simple class that holds a list of WardInfo objects, and provides
       useful search functions over that list. This prevents me from
       cluttering up the interface of Network
    """
    #: The list of WardInfo objects, one for each ward in order
    wards: _List[WardInfo] = _field(default_factory=list)

    def __len__(self):
        return len(self.wards)

    def __getitem__(self, index):
        return self.wards[index]

    def _find_ward(self, name: str, match: bool, include_alternates: bool):
        """Internal function that flexibly finds a ward by name"""
        import re

        if not isinstance(name, re.Pattern):
            search = re.compile(name, re.IGNORECASE)
        else:
            search = name

        if match:
            search = search.match
        else:
            search = search.search

        matches = []

        for i, ward in enumerate(self.wards):
            if ward is None:
                continue

            is_match = False

            if search(ward.name):
                is_match = True
            elif search(ward.code):
                is_match = True
            elif include_alternates:
                for alternate in ward.alternate_names:
                    if search(alternate):
                        is_match = True
                        break

                if not is_match:
                    for alternate in ward.alternate_codes:
                        if search(alternate):
                            is_match = True
                            break

            if is_match:
                matches.append(i)

        return matches

    def _find_authority(self, name: str, match: bool):
        """Internal function that flexibly finds a ward by authority"""
        import re

        if not isinstance(name, re.Pattern):
            search = re.compile(name, re.IGNORECASE)
        else:
            search = name

        if match:
            search = search.match
        else:
            search = search.search

        matches = []

        for i, ward in enumerate(self.wards):
            if ward is None:
                continue

            is_match = False

            if search(ward.authority):
                is_match = True
            elif search(ward.authority_code):
                is_match = True

            if is_match:
                matches.append(i)

        return matches

    def _find_region(self, name: str, match: bool):
        """Internal function that flexibly finds a ward by region"""
        import re

        if not isinstance(name, re.Pattern):
            search = re.compile(name, re.IGNORECASE)
        else:
            search = name

        if match:
            search = search.match
        else:
            search = search.search

        matches = []

        for i, ward in enumerate(self.wards):
            if ward is None:
                continue

            is_match = False

            if search(ward.region):
                is_match = True
            elif search(ward.region_code):
                is_match = True

            if is_match:
                matches.append(i)

        return matches

    def _intersect(self, list1, list2):
        """Return the intersection of two lists"""
        return [value for value in list1 if value in list2]

    def find(self, name: str = None,
             authority: str = None, region: str = None,
             match: bool = False, match_authority_and_region: bool = False,
             include_alternates: bool = True):
        """Generic search function that will search using any or all
           of the terms provided. This returns a list of indicies
           of wards that match the search

           Parameters
           ----------
           name: str or regexp
             Name or code of the ward to search
           authority: str or regexp
             Name or code of the authority to search
           region: str or regexp
             Name or code of the region to search
           match: bool (False)
             Use a regular expression match for the ward rather than a
             search. This forces the match to be at the start of the string
           match_authority_and_region: bool (False)
             Use a regular expression match for the authority and region
             rather than a search. This forces the match to be at the start
             of the string
           include_alternates: bool (True)
             Whether or not to include alternative names and codes when
             searching for the ward
        """
        wards = None

        if name is not None:
            wards = self._find_ward(name, match=match,
                                    include_alternates=include_alternates)

            if len(wards) == 0:
                return wards

        if authority is not None:
            authorities = self._find_authority(
                                        authority,
                                        match=match_authority_and_region)

            if len(authorities) == 0:
                return authorities

            if wards is None:
                wards = authorities
            else:
                wards = self._intersect(wards, authorities)
                wards.sort()
                if len(wards) == 0:
                    return wards

        if region is not None:
            regions = self._find_region(region,
                                        match=match_authority_and_region)

            if len(regions) == 0:
                return regions

            if wards is None:
                wards = regions
            else:
                wards = self._intersect(wards, regions)
                wards.sort()

        if wards is None:
            # we have not searched for anything, so return everything
            return list(range(1, len(self.wards)))
        else:
            return wards
